// INISTRUC.H - INI file data structure for 4xxx / TCMD products
//   Copyright 1997  JP Software Inc.  All rights reserved
// Copyright 2003, Doug Lindquist

// INI file constants
#define INI_DEFKEYS 4
#define INI_SIG 0x4DD4
#define INI_STRMAX 1536
#define INI_KEYMAX 64
#define INI_KEYS_SIZE (2 * sizeof(UINT) * INI_KEYMAX)
#define INI_PRODUCT 0x80
#define INI_PRIMARY 0x40
#define INI_SECONDARY 0x20
#define INI_TOTAL_BYTES (sizeof (INIFILE) + INI_STRMAX + INI_KEYS_SIZE)
#define INI_EMPTYSTR ((unsigned int)-1)

#define OPTBITS 384
#define OBSIZE  ((OPTBITS + 7) / 8)

// structures and macros for token lists
typedef struct
  {
  INT num_entries;
  INT entry_size;
  CHAR **elist;
  } TL_HEADER;

#define TOKEN_LIST(tl_name, tlist) \
  TL_HEADER tl_name = \
  { (sizeof(tlist) / sizeof(tlist[0])), sizeof(tlist[0]), (char **)&tlist }


// INI file data structure
// This structure is maintained in the following order:
//    Header
//    INI file data for:
//       All products
//       Character mode products
//       GUI products
//       Character/GUI product crossovers
//
// Within each section of INI file data, there is a further subdivision:
//    String pointers
//    Integers
//    Choices (2-valued)
//    Characters / bytes (includes 3+ valued choices)
//
// Within each of those sections, there is a further division:
//    INI file data, alphabetically by directive name
//    Internal data, alphabetically by structure element name
//

typedef struct
  {
  /*
  **********************************************************************
  ** Header (all products)
  **********************************************************************
  */
  PSZ StrData;            // (internal) pointer to strings data
  UINT StrMax;            // (internal) maximum string bytes
  UINT StrUsed;           // (internal) actual string bytes
  UINT *Keys;             // (internal) pointer to key list
  UINT KeyMax;            // (internal) maximum key count
  UINT KeyUsed;           // (internal) actual key count
  UINT SecFlag;           // (internal) INI file section bit flags
  UINT OBCount;           // (internal) OPTION bit flag count
  UCHAR OptBits[OBSIZE];  // (internal) OPTION bit flags

  /*
  **********************************************************************
  ** All Products
  **********************************************************************
  */
          // String pointers
  UINT DescriptName;      // alternative name for DESCRIPT.ION
  UINT DirColor;          // DirColors
  UINT FC;                // Filename completion
  UINT FSPath;            // 4StartPath / TCStartPath
  UINT HistLogName;       // HistoryLogName
  UINT LogName;           // LogName
  UINT Printer;           // Printer
  UINT TreePath;          // CD / CDD tree index file

          // Integers
  UINT AliasSize;         // Alias
  UINT AliasNew;          // Alias (new from OPTION, in 4DOS)
  UINT BeepDur;           // BeepLength
  UINT BeepFreq;          // BeepFreq
  UINT CDDHeight;         // CDDWinHeight
  UINT CDDLeft;           // CDDWinLeft
  UINT CDDTop;            // CDDWinTop
  UINT CDDWidth;          // CDDWinWidth
  INT CursI;                      // CursorIns
  INT CursO;                      // CursorOver
  UINT DescriptMax;       // DescriptionMax
  UINT DirHistorySize;    // DirHistory
  UINT DirHistoryNew;     // DirHistory (new from OPTION, in 4DOS)
  UINT EnvSize;           // Environment
  UINT EnvNew;                    // Environment (new from OPTION, in 4DOS)
  UINT EvalMax;           // Maximum # of EVAL decimal places
  UINT EvalMin;           // Minimum # of EVAL decimal places
  UINT FuzzyCD;           // fuzzy CD completion style
  UINT HistMin;           // HistMin
  UINT HistorySize;       // History
  UINT HistoryNew;        // History (new from OPTION, in 4DOS)
  UINT InputColor;        // color used for command input
  UINT ListColor;         // ListColors
  UINT ListRowStart;
  UINT PWHeight;          // PopupWinHeight
  UINT PWLeft;            // PopupWinLeft
  UINT PWTop;             // PopupWinTop
  UINT PWWidth;           // PopupWinWidth
  UINT SelectColor;       // SelectColors
  UINT StdColor;          // StdColors
  UINT Tabs;              // tabstops (for LIST)

  UINT INIDebug;          // (internal) debug bit flags
  INT ShellLevel;         // (internal) -1 if /C, 0 if /P, 1 otherwise
  INT ShellNum;           // (internal) shell number

          // Choices
  UCHAR AppendDir;        // Append backslash to directory on tab
  UCHAR BatEcho;          // BatchEcho
  UCHAR DecimalChar;      // decimal character for @EVAL
  UCHAR Descriptions;     // Descriptions
  UCHAR EditMode;         // EditMode
  UCHAR HistoryCopy;      // HistCopy
  UCHAR HistoryMove;      // HistMove
  UCHAR HistoryWrap;      // HistWrap
  UCHAR INIQuery;         // INIQuery
  UCHAR NoClobber;        // NoClobber
  UCHAR PathExt;          // PathExt
  UCHAR PauseErr;         // PauseOnError
  UCHAR ThousandsChar;    // decimal character for @EVAL
  UCHAR TimeFmt;          // AmPm
  UCHAR UnixPaths;        // UnixPaths
  UCHAR Upper;            // UpperCase

          // Characters / bytes
  UCHAR CmdSep;           // CommandSep
  UCHAR EscChr;           // EscapeChar
  UCHAR ParamChr;         // ParameterChar

  UCHAR BootDrive;        // (internal) boot drive letter (upper case)
  UCHAR Expansion;        // (internal) alias/variable expansion flag
  UCHAR LogOn;            // (internal) command logging flag
  UCHAR HistLogOn;        // (internal) history logging flag
  UCHAR SingleStep;       // (internal) batch file single step flag
  UCHAR SwChr;            // (internal) SwitchChar


  /*
  **********************************************************************
  ** Character Mode Products
  **********************************************************************
  */
  // ---------------------------------
  // -- All Character Mode Products
  // ---------------------------------
          // Integers
  UINT CDDColor;          // CDDWinColor
  UINT LBBar;             // ListboxBarColors
  UINT ListStatusColor;   // ListStatusColors
  UINT PWColor;           // PopupWinColor
  UINT SelectStatusColor; // SelectStatusColors

          // Choices
  UCHAR BrightBG;         // BrightBG
  UCHAR LineIn;           // LineInput

  /*
  **********************************************************************
  ** Special
  **********************************************************************
  */

          // Integers
  UINT WindowHeight;      // height, in pels
  UINT WindowState;       // WindowState
  UINT WindowWidth;       // width, in pels
  UINT WindowX;           // top left, in pels
  UINT WindowY;

          // Choices
  UCHAR ExecWait;         // if !=0, wait for apps at cmd line

          // String pointers
  UINT NextININame;       // NextINIFile
  UINT PrimaryININame;    // (internal) primary INI file name

          // Choices
  UCHAR DupBugs;          // duplicate CMD.EXE bugs?
  UCHAR LocalAliases;     // LocalAliases
  UCHAR LocalDirHistory;  // LocalDirHistory
  UCHAR LocalHistory;     // LocalHistory

  UCHAR ANSI;             // ANSI escapes enabled/disabled

          // String pointers
  UINT HelpBook;          // HelpBook
  UINT SwapPath;          // path to swap file

          // Integers
  UINT Rows;              // ScreenRows

  /*
  **********************************************************************
  ** Trailer (all products)
  **********************************************************************
  */
  UINT INIBuild;          // (internal) current internal build number
  UINT INISig;            // (internal) INI file signature
  } INIFILE;

