// LISTALL.C - file listing routines for 4xxx / TCMD family
//   Copyright (c) 1993 - 1998  Rex C. Conn  All rights reserved
// (c) 2003 Doug Lindquist

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <fcntl.h>
#include <io.h>
#include <share.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "4all.h"
#include "product.h"

static int ListInit(void);
static int ListOpenFile(char *);
static void ListHome(void);
static int ListMoveLine(int);
static LONG MoveViewPtr(LONG, LONG *);
static LONG ComputeLines(LONG,LONG);
static VOID SetRightMargin(void);
static INT DisplayLine(UINT, LONG);
int FormatLine(char [], LONG, int *, int);
static BOOL SearchFile(long);
VOID ListPrintFile(void);
void ListSetCurrent(LONG);
static void ListFileRead(PCH, unsigned int);
static int GetPrevChar(void);
static int GetNextChar(void);

LISTFILESTRUCT LFile;

CHAR  szListFindWhat[80] = "";
CHAR  szFFindFindWhat[128] = "";
static char szClip[MAXFILENAME];

static BOOL bListSkipLine = 0;
BOOL  fEndSearch = 0;

static int nScreenRows;     // screen size in rows
static int nScreenColumns;  // screen size in columns
static int nRightMargin;    // right margin (column #)
static int nNormal;         // normal video attribute
static int nInverse;        // inverse video attribute
static int nStart;          // start index & current index for
static int nCurrent;        // "previous file" support
long uListFlags;            // LIST options (/H, /S, /W, /X)

extern RANGES aRanges;

int ListEntry(char *);
static int _list(char *);
static void ListUpdateScreen(void);
static void list_wait(char *);
static void clear_header(void);
static void wWriteListStr(int, int, POPWINDOWPTR, char *);

static char dirty_header;               // if != 0, redisplay header
static long fSearchFlags = 0;

// FFIND structure
typedef struct
  {
  unsigned long fFlags;
  unsigned int uMode;
  long lFilesFound;
  long lLinesFound;
  char szDrives[32];
  char szSource[MAXFILENAME];
  char szFilename[MAXFILENAME];
  char szText[512];
  } FFIND_STRUCT;

static int _ffind(FFIND_STRUCT *);

/************************************************************************/
// find files or text
int ffind_cmd(int argc, char **argv)
  {
  char *arg, *ptr;
  int i, n, nLength, rval = 0;
  FFIND_STRUCT Find;
  LISTFILESTRUCT SaveLFile;

  // initialize variables for SearchDirectory()
  init_dir();

  memset(&Find, '\0', sizeof(FFIND_STRUCT));
  fEndSearch = 0;
  uListFlags = 0L;

  // turn off the error reporting in SearchDirectory()
  glDirFlags |= DIRFLAGS_RECURSE;

  // no "." or ".."
  glDirFlags |= DIRFLAGS_NO_DOTS;

  sprintf(Find.szDrives, FMT_CHAR, gcdisk(NULL) + 64);

  // get date/time/size ranges
  if (GetRange(argv[1], &aRanges, 1) != 0)
  return ERROR_EXIT;

  init_page_size();

  // check for and remove switches; abort if no filename arguments
  argc = 0;
  arg = ntharg(argv[1], argc | 0x1000);
  while ((arg != NULL) && (*arg == gpIniptr->SwChr))
    {
    // point past switch character
    for (arg++; (*arg != '\0');)
      {
      switch (_ctoupper(*arg++))
        {
        case 'A':
          // retrieve based on specified atts
          arg = GetSearchAttributes(arg);
          Find.uMode = 0x417;
          break;

        case 'B':   // bare display (filenames only)
          glDirFlags |= (DIRFLAGS_NAMES_ONLY | DIRFLAGS_NO_HEADER
                         | DIRFLAGS_NO_FOOTER);
          break;

        case 'C':   // case-sensitive
          Find.fFlags |= FFIND_CHECK_CASE;
          break;

        case 'D':   // search entire disk(s)
          Find.fFlags |= (FFIND_SUBDIRS | FFIND_ALL);

          // check for additional drives
          if (*arg)
            {
            sscanf(arg, "%31s%n", Find.szDrives, &i);
            arg += i;
            strupr(Find.szDrives);
            }

          break;

        case 'E':   // display filenames in upper case
          glDirFlags |= DIRFLAGS_UPPER_CASE;
          break;

        case 'I':
          Find.fFlags |= FFIND_NOWILDCARDS;
          break;

        case 'K':   // no header
          glDirFlags |= DIRFLAGS_NO_HEADER;
          break;

        case 'L':   // display line numbers
          Find.fFlags |= FFIND_LINE_NUMBERS;
          break;

        case 'M':   // no footer
          glDirFlags |= DIRFLAGS_NO_FOOTER;
          break;

        case 'O':   // dir sort order
          // kludges for DOS 5 format
          arg = dir_sort_order(arg);
          break;

        case 'P':   // pause after each page
          gnPageLength = GetScrRows();
          break;

        case 'R':   // reverse search (from end)
          Find.fFlags |= FFIND_REVERSE_SEARCH;
          break;

        case 'S':   // search subdirectories
          Find.fFlags |= FFIND_SUBDIRS;
          break;

        case 'V':   // show all matching lines / offsets
          Find.fFlags |= FFIND_SHOWALL;
          break;

        case 'X':   // hex search
          Find.fFlags |= FFIND_HEX_DISPLAY;
          if (_ctoupper(*arg) == 'T')
            arg++;
          else
            {
            if (*arg == '\0')
              break;

            Find.fFlags |= FFIND_HEX_SEARCH;
            }

        case 'T':   // text search
          if (*arg != '\0')
            {
            if (*arg == '"')
              sscanf(++arg, "%127[^\"]", szFFindFindWhat);
            else
              strcpy(szFFindFindWhat, arg);

            arg = NULLSTR;
            Find.fFlags |= FFIND_TEXT;
            }

          break;

        default:    // invalid option
          error(ERROR_INVALID_PARAMETER, arg-1);
          return (usage(FFIND_USAGE));
        }  // switch (_ctoupper(*arg++))
      }  // for (arg++; (*arg != '\0');)

    argc++;
    arg = ntharg(argv[1], argc | 0x1000);
    }

  // skip the switch statements
  if (gpNthptr == NULL)
    {
    if (QueryIsPipeHandle(STDIN))
      argv[1] = CONSOLE;
    else
      return (usage(FFIND_USAGE));
    }
  else
    strcpy(argv[1], gpNthptr);

  Find.fFlags |= FFIND_NOERROR;
  if (Find.fFlags & FFIND_TEXT)
    {
    // save the LIST variables
    memmove(&SaveLFile, &LFile, sizeof(LISTFILESTRUCT));

    // allocate buffer if searching for text
    ListInit();

    // don't search for directories for text!
    Find.uMode &= 0x407;
    }

  if (setjmp(cv.env) == -1)
    {
    rval = CTRLC;
    goto ffind_bye;
    }

  for (argc = 0; (fEndSearch == 0); argc++)
    {
    for (n = 0; ((Find.szDrives[n] != '\0') && (fEndSearch == 0)); n++)
      {
      if ((arg = ntharg(argv[1], argc)) == NULL)
        {
        fEndSearch = 1;
        break;
        }

      // no multi-drive specs if drive already named!
      if ((Find.fFlags & FFIND_ALL) && (arg[1] != ':') && (arg[1] != '\\'))
        {
        // check for drive range (c-f)
        if (Find.szDrives[n] == '-')
          {
          if ((n > 0)
              && (Find.szDrives[n + 1] > (char)(Find.szDrives[n - 1] + 1)))
            Find.szDrives[--n] += 1;
          else
            n++;
          }

        i = Find.szDrives[n];
        strins(arg, " :\\");
        *arg = (char)i;
        }
      else
        n = strlen(Find.szDrives) - 1;

      // build the source name
      if ((QueryIsDevice(arg)) && (QueryIsCON(arg)))
        uListFlags |= LIST_STDIN;
      else
        {
        if (mkfname(arg, 0) == NULL)
          {
          rval = ERROR_EXIT;
          goto ffind_bye;
          }
        else
          {
          if (glDirFlags & DIRFLAGS_UPPER_CASE)
          strupr(arg);

          // check to see if it's an HPFS partition
          if (ifs_type(arg) != 0)
          glDirFlags |= DIRFLAGS_HPFS;

          // if a directory, append "\*.*"
          // if not a directory, append a ".*" to files with no extension
          if (is_dir(arg))
            mkdirname(arg, ((glDirFlags & DIRFLAGS_HPFS) ? "*" : WILD_FILE));
          // don't do this stuff on an LFN / NTFS volume!
          else
            {
            if ((glDirFlags & DIRFLAGS_HPFS) == 0)
              {
              // if no include list & no extension specified, add default one
              //   else if no filename, insert a wildcard filename
              if (strchr(arg, ';') == NULL)
                {
                if (ext_part(arg) == NULL)
                  {
                  if (strpbrk(arg, WILD_CHARS) == NULL)
                    strcat(arg, WILD_FILE);
                  }
                else
                  {
                  // point to the beginning of the filename
                  ptr = arg + strlen(path_part(arg));
                  if (*ptr == '.')
                    strins(ptr, "*");
                  }
                }  // if (strchr(arg, ';') == NULL)
              }  // if ((glDirFlags & DIRFLAGS_HPFS) == 0)
            }
          }
        }

      copy_filename(Find.szSource, arg);

      // save the source filename part (for recursive calls &
      //   include lists)
      nLength = 0;
      arg = path_part(Find.szSource);
      if (arg != NULL)
        nLength = strlen(arg);

      strcpy(Find.szFilename, Find.szSource + nLength);

      i = _ffind(&Find);

      if ((setjmp(cv.env) == -1) || (i == CTRLC))
        {
        rval = CTRLC;
        fEndSearch = 1;
        break;
        }

      EnableSignals();
      if (i != 0)
        rval = ERROR_EXIT;
      }  // for (n = 0; ...
    }  // for (argc = 0; (fEndSearch == 0); argc++)

  if ((glDirFlags & DIRFLAGS_NO_FOOTER) == 0)
    {
    // display number of lines & files found
    crlf();
    _page_break();
    if (Find.fFlags & FFIND_TEXT)
      printf(FFIND_TEXT_FOUND, Find.lLinesFound,
             ((Find.lLinesFound == 1) ? FFIND_ONE_LINE : FFIND_MANY_LINES));

    FilesProcessed(FFIND_FOUND, Find.lFilesFound);
    crlf();
    _page_break();
    }

ffind_bye:
  if (Find.fFlags & FFIND_TEXT)
    {
    FreeMem((char *)LFile.lpBufferStart);
    if (LFile.hHandle > 0)
      _close(LFile.hHandle);

    LFile.hHandle = 0;

    // restore the LIST variables
    memmove(&LFile, &SaveLFile, sizeof(LISTFILESTRUCT));
    }

  // disable signal handling momentarily
  HoldSignals();

  return ((Find.lFilesFound != 0L) ? rval : ERROR_EXIT);
  }
/************************************************************************/
static int _ffind(FFIND_STRUCT *Find)
  {
  unsigned int i;
  char *arg;
  int c, n = 0, rval = 0, nLength;
  unsigned int entries = 0;
  long lLines;
  DIR_ENTRY *files = 0L;    // file array in system memory

  // trap ^C and clean up
  if (setjmp(cv.env) == -1)
    {
    dir_free(files);
    return CTRLC;
    }

  EnableSignals();
  if (glDirFlags & DIRFLAGS_UPPER_CASE)
    strupr(Find->szSource);
  else
    {
    if ((glDirFlags & DIRFLAGS_HPFS) == 0)
      strlwr(Find->szSource);
    }

  // look for matches
  if (((uListFlags & LIST_STDIN) == 0)
      && (SearchDirectory((Find->uMode | FIND_DATERANGE), Find->szSource,
                          (DIR_ENTRY **)&files, &entries, &aRanges, 0) != 0))
    {
    rval = ERROR_EXIT;
    }
  else
    {
    if (uListFlags & LIST_STDIN)
      entries = 1;
    else
      arg = path_part(Find->szSource);

    for (i = 0; ((i < entries) && (fEndSearch == 0)); i++)
      {
      if (uListFlags & LIST_STDIN)
        strcpy(LFile.szName, "CON");
      else
        sprintf(LFile.szName, "%s%Fs", arg,
                ((glDirFlags & DIRFLAGS_HPFS) ? files[i].hpfs_name
                                              : files[i].file_name));

      if (Find->fFlags & FFIND_TEXT)
        {
        if (uListFlags & LIST_STDIN)
          {
          LFile.hHandle = STDIN;
          LFile.lSize = (ULONG)-1;
          }
        else
          {
          LFile.hHandle = _sopen(LFile.szName, (O_BINARY | O_RDONLY),
                                 SH_DENYNO);
          if (LFile.hHandle == -1)
            {
            rval = error(_doserrno, LFile.szName);
            continue;
            }
          }

        if (LFile.hHandle != STDIN)
          LFile.lSize = QuerySeekSize(LFile.hHandle);

        LFile.lpEOF = 0L;

        ListHome();
        lLines = 0L;
        if (Find->fFlags & FFIND_REVERSE_SEARCH)
          {
          // goto the last row
          while (ListMoveLine(1) != 0);

          // display filename
          if (glDirFlags & DIRFLAGS_NAMES_ONLY)
            {
            // if /B, _only_ display name!
            more_page(LFile.szName, 0);
            break;
            }

          if ((n == 0) && ((glDirFlags & DIRFLAGS_NO_HEADER) == 0))
            {
            crlf();
            _page_break();
            sprintf(Find->szText, "---- %s", LFile.szName);
            more_page(Find->szText, 0);
            n++;
            }

          if (Find->fFlags & FFIND_HEX_DISPLAY)
            {
            sprintf(Find->szText, FFIND_OFFSET, LFile.lViewPtr,
                    LFile.lViewPtr);
            more_page(Find->szText, 0);
            }
          else
            {
            // display line number
            nLength = 0;
            if (Find->fFlags & FFIND_LINE_NUMBERS)
              nLength = sprintf(Find->szText, "[%ld] ",
                                LFile.lCurrentLine + gpIniptr->ListRowStart);

            // display the line
            c = GetNextChar();
            while ((c != EOF) && (c != LF) && (c != CR) && (nLength < 511))
              {
              if (c == 0)
                c = ' ';

              Find->szText[ nLength++ ] = (char)c;
              }

            Find->szText[ nLength ] = '\0';
            more_page(Find->szText, 0);
            c = GetNextChar();
            }

          if ((Find->fFlags & FFIND_SHOWALL) == 0)
            break;

          // skip to next line
          if ((Find->fFlags & FFIND_REVERSE_SEARCH) == 0)
            {
            if (Find->fFlags & FFIND_HEX_DISPLAY)
              LFile.lViewPtr += LFile.nSearchLen;
            else
              {
              if (ListMoveLine(1) == 0)
                break;
              }
            }

          ListSetCurrent(LFile.lViewPtr);
          }  // if (Find->fFlags & FFIND_TEXT)

        _close(LFile.hHandle);
        LFile.hHandle = 0;
        if (lLines == 0L)
          continue;

        Find->lLinesFound += lLines;
        }
      else
        {
        more_page(LFile.szName, 0);
        }

      Find->lFilesFound++;
      }  // for (i = 0; ((i < entries) && (fEndSearch == 0)); i++)

    dir_free(files);
    files = 0L;

    // search subdirectories too?
    if (Find->fFlags & FFIND_SUBDIRS)
      {
      insert_path(Find->szSource, WILD_FILE, path_part(Find->szSource));

      // save the current source filename start
      arg = strchr(Find->szSource, '*');
      entries = 0;

      // search for subdirectories
      if (SearchDirectory(((Find->uMode & 0xF) | 0x210), Find->szSource,
                          (DIR_ENTRY **)&files, &entries, &aRanges,
                          ((gszSortSequence[0] == '\0') ? 2 : 0)) != 0)
        {
        rval = ERROR_EXIT;
        goto find_abort;
        }

      for (i = 0; ((i < entries) && (fEndSearch == 0)); i++)
        {
        // make directory name
        sprintf(arg, "%Fs\\%s",
                ((glDirFlags & DIRFLAGS_HPFS) ? files[i].hpfs_name
                                   : files[i].file_name), Find->szFilename);

        rval = _ffind(Find);

        // check for ^C and reset ^C trapping
        if ((setjmp(cv.env) == -1) || (rval == CTRLC))
          rval = CTRLC;

        // quit on error (probably out of memory)
        if (rval)
          break;

        EnableSignals();
        }  // for (i = 0; ((i < entries) && (fEndSearch == 0)); i++)

      dir_free(files);
      files = 0L;
      }  // if (Find->fFlags & FFIND_SUBDIRS)
    }  // else

find_abort:
  // disable signal handling momentarily
  HoldSignals();
  return rval;
  }
/************************************************************************/
// initialize the LIST buffers & globals
static int ListInit(void)
  {
  // allocate 256K for 4OS2
  LFile.ulTotalSize = 0x40000L;

  LFile.lpBufferStart = AllocMem(LFile.ulTotalSize);
  if (LFile.lpBufferStart == 0L)
    return (error(ERROR_NOT_ENOUGH_MEMORY, NULL));

  LFile.ulBufferSize = (LFile.ulTotalSize / 2);
  LFile.lpBufferEnd = LFile.lpBufferStart + LFile.ulTotalSize;
  LFile.lpCurrent = LFile.lpBufferStart;
  LFile.lpEOF = 0L;

  // initialize LISTFILESTRUCT parms
  LFile.szName[0] = '\0';
  LFile.hHandle = 0;
  LFile.lSize = LFile.lCurrentLine = 0L;

  LFile.lViewPtr = LFile.lFileOffset = 0L;

  // set screen sizes
  nScreenRows = GetScrRows();
  nScreenColumns = GetScrCols();
  nScreenColumns--;
  return 0;
  }
/************************************************************************/
// open the file & initialize buffers
static int ListOpenFile(char *fname)
  {
  LFile.lpEOF = 0L;

  if (uListFlags & LIST_STDIN)
    {
    // reading from STDIN
    LFile.hHandle = STDIN;
    strcpy(LFile.szName, LIST_STDIN_MSG);
    LFile.lSize = (ULONG)-1;
    }
  else
    {
    if (LFile.hHandle > 0)
      _close(LFile.hHandle);

    StripQuotes(fname);
    LFile.hHandle = _sopen(((szClip[0]) ? szClip : fname),
                           (O_BINARY | O_RDONLY), SH_DENYNO);
    if (LFile.hHandle == -1)
      {
      return (error(_doserrno, fname));
      }

    LFile.lSize = QuerySeekSize(LFile.hHandle);
    strcpy(LFile.szName, fname);
    }

  ListHome();
  ListUpdateScreen();
  return 0;
  }
/************************************************************************/
// reset to the beginning of the file
static void ListHome(void)
  {
  // Reset the memory buffer to purge old file data
  LFile.lpBufferEnd = LFile.lpBufferStart + LFile.ulTotalSize;
  LFile.lpCurrent = LFile.lpBufferStart;
  LFile.lViewPtr = LFile.lFileOffset = 0L;
  LFile.lCurrentLine = 0L;
  LFile.nListHorizOffset = 0;

  ListFileRead(LFile.lpBufferStart, LFile.ulTotalSize);

  // determine EOL character for this file
  for (LFile.fEoL = CR; ((LFile.lpCurrent != LFile.lpEOF) && (LFile.lpCurrent < LFile.lpBufferEnd)); LFile.lpCurrent++)
    {
    if (*LFile.lpCurrent == LF)
      {
      LFile.fEoL = LF;
      break;
      }
    }

  LFile.lpCurrent = LFile.lpBufferStart;
  }
/************************************************************************/
// move up/down "n" lines
static int ListMoveLine(int nRows)
  {
  long lTemp, lRows;

  lRows = nRows;
  lTemp = MoveViewPtr(LFile.lViewPtr, &lRows);

  if (lRows == nRows)
    {
    LFile.lCurrentLine += lRows;
    LFile.lViewPtr += lTemp;
    return 1;
    }

  return 0;
  }
/************************************************************************/
static VOID SetRightMargin(void)
  {
  nRightMargin = ((uListFlags & LIST_WRAP) ? nScreenColumns + 1 : 511);
  }
/************************************************************************/
// read the current file into the specified buffer
static void ListFileRead(PCH fptr, unsigned int size)
  {
  ULONG ulBytesRead = 0;

  // position the read pointer
  if (LFile.hHandle == STDIN)
    {
    if (LFile.lpEOF != 0L)
      return;

    // read until buffer full or no more data
    do
      {
      if (FileRead(LFile.hHandle, fptr, size, &ulBytesRead) != 0)
        break;

      fptr += ulBytesRead;
      } while ((ulBytesRead > 0) && ((size -= ulBytesRead) > 0));

    if (size == 0)
      ulBytesRead = 0;

    // are we at the end of the file yet?
    if (ulBytesRead == size)
      LFile.lpEOF = 0L;
    else
      {
      LFile.lpEOF = (fptr + ulBytesRead);
      LFile.lSize = (long)(fptr - (LONG)LFile.lpBufferStart)
                                   + LFile.lFileOffset;
      }

    return;
    }  // if (LFile.hHandle == STDIN)

  _lseeki64(LFile.hHandle, LFile.lFileOffset, SEEK_SET);
  FileRead(LFile.hHandle, fptr, size, &ulBytesRead);

  // check for file length the same size as the input buffer!
  LFile.lpEOF = (((LFile.lSize > (LONG)(LFile.lFileOffset + ulBytesRead))
                 && (ulBytesRead == size)) ? 0L : fptr + ulBytesRead);
  }
/************************************************************************/
// set LFile.lpCurrent to the specified file offset
void ListSetCurrent(long lOffset)
  {
  if (lOffset < LFile.lFileOffset)
    {
    // get previous block
    while ((lOffset < LFile.lFileOffset) && (LFile.lFileOffset > 0L))
      {
      if ((LFile.lFileOffset -= LFile.ulTotalSize) < 0L)
        LFile.lFileOffset = 0L;
      }

    ListFileRead(LFile.lpBufferStart, LFile.ulTotalSize);
    }
  else
    {
    if (lOffset > (long)(LFile.lFileOffset + LFile.ulTotalSize))
      {
      // get next block
      while (lOffset > (long)(LFile.lFileOffset + LFile.ulTotalSize))
        LFile.lFileOffset += LFile.ulTotalSize;

      ListFileRead(LFile.lpBufferStart, LFile.ulTotalSize);
      }
    }

  // it's (now) in the current buffer
  LFile.lpCurrent = LFile.lpBufferStart + (lOffset - LFile.lFileOffset);
  }
/************************************************************************/
// Display Line
static INT DisplayLine(UINT nRow, LONG lLinePtr)
  {
  int i, n;
  INT nLength, nBytesPrinted = 0, nHOffset = 0, nHexOffset;
  CHAR *arg, szBuffer[512], cSave;

  ListSetCurrent(lLinePtr);

  nLength = FormatLine(szBuffer, lLinePtr, &nBytesPrinted, TRUE);

  if ((uListFlags & LIST_HEX) == 0)
    {
    nHOffset = LFile.nListHorizOffset;
    nLength -= LFile.nListHorizOffset;
    nHexOffset = 0;
    }
  else
    nHexOffset = 9;

  // adjust to max screen width
  if (nLength > (nScreenColumns + 1))
    nLength = nScreenColumns + 1;

  if ((nBytesPrinted > 0) && (nLength > 0))
    {
    VioWrtCharStrAtt(szBuffer+nHOffset, nLength, nRow, 0, (PBYTE)&nNormal, 0);
    }

  // if we're displaying a search result, highlight the string
  if ((LFile.fDisplaySearch) && (strlen(szBuffer) > (unsigned int)nHOffset))
    {
    for (i = 0; ;)
      {
      if (LFile.fDisplaySearch & 2)
        arg = strstr(szBuffer + nHOffset + i + nHexOffset, szListFindWhat);
      else
        arg = stristr(szBuffer + nHOffset + i + nHexOffset, szListFindWhat);

      if (arg == NULL)
        break;

      n = strlen(szListFindWhat);
      cSave = arg[ n ];
      arg[ n ] = '\0';
      i = (int)(arg - (szBuffer + nHOffset));
      WriteStrAtt(nRow, i, nInverse, arg);
      arg[ n ] = cSave;
      i += n;
      }  // for
    }  // if ((LFile.fDisplaySearch)...

  return nBytesPrinted;
  }
/************************************************************************/
// Compute number of lines between the specified points in the file
static LONG ComputeLines(LONG lOldPtr, LONG lNewPtr)
  {
  int ch, n, fDirection = 0;
  LONG lRowCount;

  // always count upwards (it makes wrap computations a lot easier!)
  if (lOldPtr > lNewPtr)
    {
    lRowCount = lOldPtr;
    lOldPtr = lNewPtr;
    lNewPtr = lRowCount;
    fDirection++;
    }

  lRowCount = 0L;
  ListSetCurrent(lOldPtr);

  if ((uListFlags & LIST_HEX) == 0)
    {
    SetRightMargin();

    for (n = 0; (lOldPtr < lNewPtr); lOldPtr++)
      {
      // don't call GetNextChar unless absolutely necessary!
      if (LFile.lpCurrent == LFile.lpEOF)
        break;

      if (LFile.lpCurrent < LFile.lpBufferEnd)
        ch = *LFile.lpCurrent++;
      else
        {
        if ((ch = GetNextChar()) == EOF)
          break;
        }

      // do tab expansion
      if (ch == TAB)
        n = (((n + TABSIZE) / TABSIZE) * TABSIZE);
      else
        {
        if ((++n >= nRightMargin) || (ch == LFile.fEoL))
          {
          // if EOL increment the row count
          lRowCount++;
          n = 0;
          }
        }
      }  // for (n = 0; (lOldPtr < lNewPtr); lOldPtr++)
    }  // if ((uListFlags & LIST_HEX) == 0)
  else
    lRowCount = ((lNewPtr - lOldPtr) / 16);

  return ((fDirection) ? -lRowCount : lRowCount);
  }
/************************************************************************/
// format the current line in either ASCII or hex mode
int FormatLine(char szBuffer[], long lLinePtr, int *nBytesPrinted, int fTabs)
  {
  int i, n;
  char szScratch[4];

  *nBytesPrinted = 0;

  if ((uListFlags & LIST_HEX) == 0)
    {
    // Print as ASCII until CR/LF or for "nRightMargin" characters.

    SetRightMargin();
    for (i = 0; (i < 511);)
      {
      // don't call GetNextChar unless absolutely necessary
      if (LFile.lpCurrent == LFile.lpEOF)
        break;

      if (LFile.lpCurrent < LFile.lpBufferEnd)
        szBuffer[i] = *LFile.lpCurrent++;
      else
        szBuffer[i] = (char)GetNextChar();

      (*nBytesPrinted)++;

      if (szBuffer[i] == (char)LFile.fEoL)
        break;

      // ignore CR's
      if (szBuffer[i] == CR)
        continue;

      if (i >= nRightMargin)
        {
        GetPrevChar();
        (*nBytesPrinted)--;
        break;
        }

      // Expand Tab Characters
      if ((szBuffer[i] == TAB) && (fTabs))
        {
        n = (((i + TABSIZE) / TABSIZE) * TABSIZE);
        for (; ((i < n) && (i < nRightMargin)); i++)
          {
          // insert spaces until next tab stop
          szBuffer[i] = ' ';
          }
        }
      else
        {
        if (uListFlags & LIST_HIBIT)
          szBuffer[i] &= 0x7F;

        i++;
        }
      }  // for (i = 0; (i < 511);)

    szBuffer[i] = '\0';
    }
  else
    {
    // Convert from Hex and print.

    // clear the whole line to blanks
    sprintf(szBuffer, "%04lx %04lx%70s", (long)(lLinePtr/0x10000L),
            (long)(lLinePtr&0xFFFF), NULLSTR);

    for (i = 0; (i < 16); i++)
      {
      n = GetNextChar();
      if (n == EOF)
        break;

      if (uListFlags & LIST_HIBIT)
        n &= 0x7F;

      sprintf(szScratch, "%02x", n);
      memmove(szBuffer+10+(i*3)+(i/8), szScratch, 2);
      szBuffer[ i+61 ] = (char)((((n >= TAB) && (n <= CR)) || (n == 0)
                                  || (n == 27)) ? '.' : n);
      }  // for i

    *nBytesPrinted = i;
    i = 78;
    }  // else

  return i;
  }
/************************************************************************/
// MoveViewPtr - Update view pointer by scrolling the number of
//   lines specified in ptrRow.  If limited by the top or bottom
//   of the file, modify ptrRow to indicate the actual number of rows.
static LONG MoveViewPtr(LONG lFilePtr, LONG *ptrRow)
  {
  int i, n;
  LONG lSaveOffset, lOffset = 0L, lRow, lRowCount = 0L;
  PCH lpTemp, lpSave, lpEnd;

  ListSetCurrent(lFilePtr);

  lRow = *ptrRow;

  if ((uListFlags & LIST_HEX) == 0)
    {
    SetRightMargin();
    if (lRow > 0)
      {
      while (lRowCount < lRow)
        {
        lpSave = LFile.lpCurrent;
        lSaveOffset = LFile.lFileOffset;

        for (n = 0; ;)
          {
          i = GetNextChar();
          if ((i == EOF) || (i == LFile.fEoL))
            break;

          if (n >= nRightMargin)
            {
            GetPrevChar();
            break;
            }

          // get tab expansion
          if (i == TAB)
            n = (((n + TABSIZE) / TABSIZE) * TABSIZE);
          else
            n++;
          }

        // check for buffer move
        if (LFile.lFileOffset != lSaveOffset)
          lpSave -= LFile.ulBufferSize;

        lOffset += (long)((long)LFile.lpCurrent - (long)lpSave);

        if (i == EOF)
          break;

        lRowCount++;
        }  // while (lRowCount < lRow)
      }  // if (lRow > 0)
    else
      {
      // move backwards
      while (lRowCount > lRow)
        {
        lSaveOffset = LFile.lFileOffset;

        // get previous line
        lpEnd = LFile.lpCurrent;
        n = GetPrevChar();
        i = GetPrevChar();
        while ((i != EOF) && (i != LFile.fEoL))
          i = GetPrevChar();

        // if not at start of file, move to beginning
        //   of the line
        if (i == EOF)
          {
          if (lpEnd == LFile.lpCurrent)
            break;
          }
        else
          GetNextChar();

        lpTemp = lpSave = LFile.lpCurrent;

        // check for buffer move
        if (LFile.lFileOffset != lSaveOffset)
          lpEnd += LFile.ulBufferSize;

        if (lpEnd > LFile.lpBufferEnd)
          lpEnd = LFile.lpBufferEnd;

        // adjust for long or wrapped lines
        for (n = 0; (lpTemp < lpEnd); lpTemp++)
          {
          if (*lpTemp != (char)LFile.fEoL)
            {
            // check if past right margin but less
            //   than current position
            if ((n >= nRightMargin) && (lpTemp + 1 < lpEnd))
              {
              n = 0;
              lpSave = lpTemp;
              }

            // kludge for TAB offsets
            if (*lpTemp == TAB)
              n = (((n + TABSIZE) / TABSIZE) * TABSIZE);
            else
              n++;
            }  // if (*lpTemp != (char)LFile.fEoL)
          }  // for (n = 0; (lpTemp < lpEnd); lpTemp++)

        lOffset += (long)((long)lpSave - (long)lpEnd);
        lRowCount--;
        }  // while (lRowCount > lRow);
      }  // else

    *ptrRow = lRowCount;
    }  // if ((uListFlags & LIST_HEX) == 0)
  else
    {
    lOffset = (lRow * 16);
    *ptrRow = lRow;

    // Limit Checking
    if (((lFilePtr == 0) && (lRow < 0))
        || ((lFilePtr == LFile.lSize) && (lRow > 0)))
      {
      lOffset = 0L;
      *ptrRow = 0;
      }
    else
      {
      if ((lFilePtr + lOffset) < 0)
        {
        lOffset = -lFilePtr;
        *ptrRow = (INT)(lOffset / 16);
        }
      else
        {
        if ((lFilePtr + lOffset) > LFile.lSize)
          {
          lOffset = (LONG)(LFile.lSize - lFilePtr);
          *ptrRow = (INT)(lOffset / 16);
          }
        }
      }
    }  // else

  return lOffset;
  }
/************************************************************************/
// SearchFile - Search the file for the value in szSearchStr.
//   If found, the file pointer is modified to the location in the file.
static BOOL SearchFile(long fFlags)
  {
  extern int wild_brackets(char *, int, int);
  int c, i;
  INT   nOffset, fIgnoreCase = 0, fWildCards = 0, fWildStar;
  UINT  uLimit;
  LONG  lTemp, lLocalPtr;
  CHAR  szSearchStr[128], *pszSearchStr;

  lLocalPtr = ((fFlags & FFIND_TOPSEARCH) ? 0L : LFile.lViewPtr);
  ListSetCurrent(lLocalPtr);

  pszSearchStr = ((fFlags & FFIND_NOERROR)
                  ? szFFindFindWhat : szListFindWhat);

  if (fFlags & FFIND_HEX_SEARCH)
    {
    // convert ASCII hex string to binary equivalent
    strupr(pszSearchStr);
    i = 0;
    while ((sscanf(pszSearchStr, "%x%n", &uLimit, &LFile.nSearchLen) != 0)
           && (LFile.nSearchLen > 0))
      {
      sprintf(szSearchStr+i, FMT_CHAR, uLimit);
      pszSearchStr += LFile.nSearchLen;
      i++;
      }

    if (i == 0)
      return 0;

    szSearchStr[i] = '\0';
    LFile.nSearchLen = i;
    }
  else
    {
    // check for a wildcard search
    // a leading ` means "ignore wildcard characters"
    if (*pszSearchStr == '`')
      {
      pszSearchStr++;

      // strip (optional) trailing `
      i = strlen(pszSearchStr);
      if ((i > 0) && (pszSearchStr[--i] == '`'))
        pszSearchStr[i] = '\0';
      }
    else
      {
      if (((fFlags & FFIND_NOWILDCARDS) == 0)
          && (strpbrk(pszSearchStr, WILD_CHARS) != NULL))
        fWildCards = 1;
      }

    strcpy(szSearchStr, pszSearchStr);
    if ((LFile.nSearchLen = strlen(szSearchStr)) == 0)
      return 0;

    if ((fFlags & FFIND_CHECK_CASE) == 0)
      {
      strupr(szSearchStr);
      fIgnoreCase = 1;
      }
    }  // else

  if (fFlags & FFIND_REVERSE_SEARCH)
    strrev(szSearchStr);

  if (fWildCards == 0)
    {
    // not a real elegant search algorithm, but real small!
    for (i = 0; ;)
      {
      if ((fFlags & FFIND_REVERSE_SEARCH) == 0)
        {
        // don't call GetNextChar unless absolutely necessary!
        if (LFile.lpCurrent == LFile.lpEOF)
          break;

        if (LFile.lpCurrent < LFile.lpBufferEnd)
          c = *LFile.lpCurrent++;
        else
          {
          // abort search if ESC key is hit
          if ((_kbhit() != 0)
              && (GetKeystroke(EDIT_NO_ECHO | EDIT_BIOS_KEY) == ESC))
            break;

          c = GetNextChar();
          if (c == EOF)
            break;
          }
        }
      else
        {
        c = GetPrevChar();
        if (c == EOF)
          break;
        }

      if ((c >= 'a') && (fIgnoreCase))
        {
        if (c <= 'z')
          c -= 32;
        else
          {
          if (c >= 0x80)
            c = _ctoupper(c);
          }
        }

      if (c != (int)szSearchStr[i])
        {
        // rewind file pointer
        if (i > 0)
          {
          if ((fFlags & FFIND_REVERSE_SEARCH) == 0)
            LFile.lpCurrent -= i;
          else
            LFile.lpCurrent += i;

          i = 0;
          }
        }
      else
        {
        if (++i >= LFile.nSearchLen)
          break;
        }
      }  // for (i = 0; ;)
    }  // if (fWildCards == 0)
  else
    {
    // search with wildcards

RestartWildcardSearch:
    fWildStar = 0;
    nOffset = 0;
    i = 0;
    while (szSearchStr[i] != '\0')
      {
      if (szSearchStr[i] == '*')
        {
        // skip past * and advance file pointer until a match
        //   of the characters following the * is found.
        fWildStar = 1;
        while ((szSearchStr[++i] == '*') || (szSearchStr[i] == '?'));
        }
      else
        {
        if ((fFlags & FFIND_REVERSE_SEARCH) == 0)
          {
          // don't call GetNextChar unless absolutely necessary!
          if (LFile.lpCurrent == LFile.lpEOF)
            {
            i = 0;
            break;
            }

          if (LFile.lpCurrent < LFile.lpBufferEnd)
            c = *LFile.lpCurrent++;
          else
            {
            // abort search if ESC key is hit
            if ((_kbhit() != 0)
                && (GetKeystroke(EDIT_NO_ECHO | EDIT_BIOS_KEY) == ESC))
              break;

            c = GetNextChar();
            }
          }
        else
          {
          if ((c = GetPrevChar()) == EOF)
            {
            i = 0;
            break;
            }
          }

        nOffset++;

        if ((c >= 'a') && (fIgnoreCase))
          {
          if (c <= 'z')
            c -= 32;
          else
            {
            if (c >= 0x80)
              c = _ctoupper(c);
            }
          }

        // ? matches any single character
        if (szSearchStr[i] == '?')
          i++;
        else
          {
          if ((szSearchStr[i] == '[') && (fWildStar == 0))
            {
            // [ ] checks for a single character
            //   (including ranges)
            if (wild_brackets(szSearchStr + i, c, fIgnoreCase) != 0)
              goto wild_rewind;

            while ((szSearchStr[i]) && (szSearchStr[i++] != ']'));
            }
          else
            {
            if (fWildStar)
              {
              int nSaveI, nSaveOffset;

              // following a '*'; so we need to do a complex
              //   match since there could be any number of
              //   preceding characters
              nSaveI = i;
              nSaveOffset = nOffset;
              for (;;)
                {
                // don't search past the current line
                if (((c == CR) || (c == LF))
                    && ((fFlags & FFIND_HEX_SEARCH) == 0))
                  goto RestartWildcardSearch;

                if (szSearchStr[i] == '[')
                  {
                  // get the first matching char
                  if (wild_brackets(szSearchStr+i, c, fIgnoreCase) == 0)
                    {
                    while ((szSearchStr[i]) && (szSearchStr[i++] != ']'));
                    }
                  }
                else
                  if (c == (int)szSearchStr[i])
                    i++;
                  else
                    {
                    // no match yet - keep scanning
                    i = nSaveI;
                    if (--nOffset > nSaveOffset)
                      {
                      if ((fFlags & FFIND_REVERSE_SEARCH) == 0)
                        LFile.lpCurrent -= (nOffset - nSaveOffset);
                      else
                        LFile.lpCurrent += (nOffset - nSaveOffset);
                      }

                    nOffset = nSaveOffset;

                    if ((fFlags & FFIND_REVERSE_SEARCH) == 0)
                      nSaveOffset++;
                    else
                      {
                      if (nSaveOffset > 0)
                        nSaveOffset--;
                      }

                    goto NextWildChar;
                    }  // else

                  if ((szSearchStr[i] == '\0') || (szSearchStr[i] == '*')
                      || (szSearchStr[i] == '?'))
                    break;

NextWildChar:
                  if ((fFlags & FFIND_REVERSE_SEARCH) == 0)
                    {
                    // don't call GetNextChar unless
                    //   absolutely necessary!
                    if (LFile.lpCurrent == LFile.lpEOF)
                      {
                      i = 0;
                      break;
                      }

                    if (LFile.lpCurrent < LFile.lpBufferEnd)
                      c = *LFile.lpCurrent++;
                    else
                      c = GetNextChar();

                    if (c == LFile.fEoL)
                      {
                      i = 0;
                      break;
                      }
                    }
                  else
                    {
                    c = GetPrevChar();
                    if (c == EOF)
                      {
                      i = 0;
                      break;
                      }
                    }

                  nOffset++;

                  if (fIgnoreCase)
                    c = _ctoupper(c);
                  }  // for (nSaveI = i, nSaveOffset = nOffset; ;)

              // if SearchStr is still an expression, we
              //   failed to find a match
              if (szSearchStr[i] == '[')
                {
                i = 0;
                nOffset = 0;
                }

              fWildStar = 0;
              }
            else
              {
              if (c == (int)szSearchStr[i])
                i++;
              else
                {
wild_rewind:
                if (--nOffset > 0)
                  {
                  if ((fFlags & FFIND_REVERSE_SEARCH) == 0)
                    LFile.lpCurrent -= nOffset;
                  else
                    LFile.lpCurrent += nOffset;
                  }

                nOffset = 0;
                i = 0;
                }
              }  // else
            }  // else
          }
        }  // else
      }  // while (szSearchStr[i] != '\0')

    // if at the end of the search string, we got a good match
    if ((szSearchStr[i] == '\0') && (nOffset > 0))
      i = LFile.nSearchLen = nOffset;
    else
      i = 0;
    }  // else

  if (i == LFile.nSearchLen)
    {
    // get current position
    LFile.lViewPtr = (LFile.lFileOffset + (long)((long)LFile.lpCurrent
                     - (long)LFile.lpBufferStart));
    if ((fFlags & FFIND_REVERSE_SEARCH) == 0)
      LFile.lViewPtr -= LFile.nSearchLen;

    // hex search from FFIND?
    if ((fFlags & FFIND_NOERROR) && (fFlags & FFIND_HEX_DISPLAY))
        return 1;

    // adjust line start & line count for hex display
    if (uListFlags & LIST_HEX)
      {
      LFile.lViewPtr -= (LFile.lViewPtr % 16);
      LFile.lCurrentLine = (LFile.lViewPtr / 16);
      }
    else
      {
      // get beginning of current line
      i = 0;
      c = GetPrevChar();
      while ((c != EOF) && (c != LFile.fEoL))
        {
        i++;
        c = GetPrevChar();
        }

      if (c != EOF)
        GetNextChar();

      // get line number (tricky because the screen may
      //   be wrapped)
      lTemp = (LFile.lFileOffset + (long)((long)LFile.lpCurrent
               - (long)LFile.lpBufferStart));
      c = (int)ComputeLines(lTemp, LFile.lViewPtr);

      if (fFlags & FFIND_TOPSEARCH)
        LFile.lCurrentLine = 0;

      LFile.lCurrentLine += ComputeLines(lLocalPtr, lTemp);

      // get beginning of line with search text
      for (LFile.lViewPtr = lTemp; (c > 0); c--)
        ListMoveLine(1);

      ListSetCurrent(LFile.lViewPtr);
      }  // else

    return 1;
    }  // if (i == LFile.nSearchLen)

  // no matching string found
  if ((fFlags & FFIND_NOERROR) == 0)
    {
    clear_header();
    WriteStrAtt(0, 1, nInverse, LIST_NOT_FOUND);
    SetCurPos(0, strlen(LIST_NOT_FOUND) + 1);
    honk();
    GetKeystroke(EDIT_NO_ECHO | EDIT_BIOS_KEY);
    }

  return 0;
  }
/************************************************************************/
// get the previous buffer character
static int GetPrevChar(void)
  {
  if (LFile.lpCurrent <= LFile.lpBufferStart)
    {     // at top of buffer?
    // can't "unwind" a pipe!
    if ((LFile.hHandle == STDIN) || (LFile.lFileOffset == 0L))
      return EOF;

    // move the 1st half of the buffer to the 2nd half
    memmove(LFile.lpBufferStart + LFile.ulBufferSize, LFile.lpBufferStart,
            LFile.ulBufferSize);

    // read the previous block into the 1st half
    LFile.lFileOffset -= LFile.ulBufferSize;

    ListFileRead(LFile.lpBufferStart, LFile.ulBufferSize);
    LFile.lpCurrent += LFile.ulBufferSize;
    }

  return *(--LFile.lpCurrent);
  }
/************************************************************************/
// get the next buffer character
static int GetNextChar(void)
  {
  if (LFile.lpCurrent == LFile.lpEOF)       // already at file end?
    return EOF;

  if (LFile.lpCurrent >= LFile.lpBufferEnd)
    {   // at end of buffer?
    // move the 2nd half of the buffer to the 1st half
    memmove(LFile.lpBufferStart, LFile.lpBufferStart + LFile.ulBufferSize,
            LFile.ulBufferSize);

    // read a new block into the 2nd half & adjust all the
    //   pointers back to the 1st half
    LFile.lFileOffset += (unsigned int)LFile.ulTotalSize;
    ListFileRead(LFile.lpBufferStart + LFile.ulBufferSize, LFile.ulBufferSize);

    LFile.lFileOffset -= LFile.ulBufferSize;
    LFile.lpCurrent -= LFile.ulBufferSize;
    }

  return *LFile.lpCurrent++;
  }
/************************************************************************/
// dummy routine for far entry to list_cmd
int ListEntry(char *pszFilename)
  {
  char *argv[3];

  argv[1] = pszFilename;
  argv[2] = NULL;
  return (list_cmd(2, argv));
  }
/************************************************************************/
// edit or display a file with vertical & horizontal scrolling & text searching
int list_cmd(int argc, char **argv)
  {
  int fval, rval = 0;
  char szSource[MAXFILENAME], szFileName[MAXFILENAME], *arg;
  FILESEARCH dir;

  memset(&dir, '\0', sizeof(FILESEARCH));

  // check for and remove switches
  if (GetRange(argv[1], &(dir.aRanges), 0) != 0)
    return ERROR_EXIT;

  // check for /T"search string"
  GetMultiCharSwitch(argv[1], "T", szSource, 255);
  if (szSource[0] == '"')
    sscanf(szSource+1, "%79[^\"]", szListFindWhat);
  else
    {
    if (szSource[0])
      sprintf(szListFindWhat, FMT_PREC_STR, 79, szSource);
    }

  if (GetSwitches(argv[1], "*HIRSWX", &uListFlags, 0) != 0)
    return (usage(LIST_USAGE));

  if (szSource[0])
    uListFlags |= LIST_SEARCH;

  // check for pipe to LIST w/o explicit /S switch
  if (first_arg(argv[1]) == NULL)
    {
    if (QueryIsPipeHandle(STDIN))
      uListFlags |= LIST_STDIN;
    else
      {
      if ((uListFlags & LIST_STDIN) == 0)
        return (usage(LIST_USAGE));
      }
    }

  // initialize buffers & globals
  if (ListInit())
    return ERROR_EXIT;

  nCurrent = nStart = 0;

  dir.hdir = INVALID_HANDLE_VALUE;

  // ^C handling
  if (setjmp(cv.env) == -1)
    {
list_abort:
    DosFindClose(dir.hdir);
    cls_cmd(1, NULL);
    rval = CTRLC;
    goto list_bye;
    }

RestartFileSearch:
  for (argc = 0; ; argc++)
    {
    // break if at end of arg list, & not listing STDIN
    if ((arg = ntharg(argv[1], argc)) == NULL)
      {
      if ((uListFlags & LIST_STDIN) == 0)
        break;
      }
    else
      strcpy(szSource, arg);

    for (fval = FIND_FIRST; ;)
      {
      szClip[0] = '\0';

      // if not reading from STDIN, get the next matching file
      if ((uListFlags & LIST_STDIN) == 0)
        {
        // qualify filename
        mkfname(szSource, 0);
        if (is_dir(szSource))
          mkdirname(szSource, WILD_FILE);

        if (stricmp(szSource, CLIP) == 0)
          {
          RedirToClip(szClip, 99);
          if (CopyFromClipboard(szClip) != 0)
            break;

          strcpy(szFileName, szClip);
          }
        else
          {
          if (QueryIsPipeName(szSource))
            {
            // only look for pipe once
            if (fval == FIND_NEXT)
              break;

            copy_filename(szFileName, szSource);
            }
          else
            {
            if (find_file(fval, szSource,
                          (FIND_BYATTS | FIND_CREATE | FIND_DATERANGE | 0x07),
                          &dir, szFileName) == NULL)
              {
              rval = ((fval == FIND_FIRST) ? ERROR_EXIT : 0);
              break;
              }
            else
              {
              if (nStart < nCurrent)
                {
                nStart++;
                fval = FIND_NEXT;
                continue;
                }
              else
                {
                if (dir.size > 0L)
                  LFile.lSize = dir.size;
                }
              }
            }
          }
        }  // if ((uListFlags & LIST_STDIN) == 0)

      // clear the screen
      cls_cmd(1, NULL);

      if ((rval = _list(szFileName)) == CTRLC)
        goto list_abort;

      if (szClip[0])
        remove(szClip);

      if (rval != 0)
        break;

      SetCurPos(nScreenRows, 0);

      if ((szClip[0]) || (uListFlags & LIST_STDIN))
        break;

      LFile.hHandle = _close(LFile.hHandle);

      // increment index to current file
      if (nCurrent < nStart)
        {
        DosFindClose(dir.hdir);
        nStart = 0;
        goto RestartFileSearch;
        }
      else
        {
        fval = FIND_NEXT;
        nCurrent++;
        nStart++;
        }
      }  // for (fval = FIND_FIRST; ;)

    // we can only read STDIN once!
    uListFlags &= ~LIST_STDIN;
    }  // for (argc = 0; ; argc++)

  crlf();

list_bye:
  FreeMem((char *)LFile.lpBufferStart);
  if (LFile.hHandle > 0)
    _close(LFile.hHandle);

  // disable signal handling momentarily
  HoldSignals();
  return rval;
  }
/************************************************************************/
static int _list(char *pszFileName)
  {
  INT c, i;
  CHAR szDescription[512], szHeader[132], szLine[10];
  LONG lTemp, lRow;
  POPWINDOWPTR wn=NULL;
  FILESEARCH dir;

  // get default normal and inverse attributes
  if (gpIniptr->ListColor != 0)
    {
    Scroll(0, 0, nScreenRows, nScreenColumns, 0, gpIniptr->ListColor);
    }

  // set colors
  GetAtt((unsigned int *)&nNormal, (unsigned int *)&nInverse);
  if (gpIniptr->ListStatusColor != 0)
    nInverse = gpIniptr->ListStatusColor;

  // flip the first line to inverse video
  clear_header();

  // open the file & initialize buffers
  if (ListOpenFile(pszFileName))
    return ERROR_EXIT;

  // kludge for empty files or pipes
  if (LFile.lSize == 0L)
    LFile.lSize = 1L;

  for (;;)
    {
    // display header
    if (dirty_header)
      {
      clear_header();
      sprintf(szHeader, LIST_HEADER, fname_part(LFile.szName),
              gchVerticalBar, gchVerticalBar, gchVerticalBar);
      WriteStrAtt(0, 0, nInverse, szHeader);
      dirty_header = 0;
      }

    // display location within file
    //  (don't use color_printf() so we won't have
    //  problems with windowed sessions)
    i = sprintf(szHeader, LIST_LINE, LFile.nListHorizOffset,
                LFile.lCurrentLine + gpIniptr->ListRowStart,
                (int)(((LFile.lViewPtr + 1) * 100) / LFile.lSize));
    WriteStrAtt(0, (nScreenColumns - i), nInverse, szHeader);
    SetCurPos(0, 0);

    if (uListFlags & LIST_SEARCH)
      {
      uListFlags &= ~LIST_SEARCH;
      fSearchFlags = 0;
      if (uListFlags & LIST_REVERSE)
        {
        c = LIST_FIND_CHAR_REVERSE;
        fSearchFlags |= FFIND_REVERSE_SEARCH;

        // goto the last row
        while (ListMoveLine(1) != 0);
        }

      if (uListFlags & LIST_NOWILDCARDS)
        fSearchFlags |= FFIND_NOWILDCARDS;

      bListSkipLine = 0;
      goto FindNext;
      }  // if (uListFlags & LIST_SEARCH)

    // get the key from the BIOS, because
    //   STDIN might be redirected
    c = cvtkey(GetKeystroke(EDIT_NO_ECHO | EDIT_BIOS_KEY | EDIT_UC_SHIFT),
                            MAP_GEN | MAP_LIST);
    if (c == ESC)
      break;

    switch (c)
      {
      case CTRLC:
        return CTRLC;

      case CUR_LEFT:
      case CTL_LEFT:
        if ((uListFlags & LIST_WRAP) || (LFile.nListHorizOffset == 0))
          goto bad_key;

        if ((LFile.nListHorizOffset -= ((c == CUR_LEFT) ? 8 : 40)) < 0)
          LFile.nListHorizOffset = 0;

        break;

      case CUR_RIGHT:
      case CTL_RIGHT:
        if ((uListFlags & LIST_WRAP) || (LFile.nListHorizOffset == 512))
          goto bad_key;

        if ((LFile.nListHorizOffset += ((c == CUR_RIGHT) ? 8 : 40)) > 512)
          LFile.nListHorizOffset = 512;

        break;

      case CUR_UP:
        if (ListMoveLine(-1) == 0)
          goto bad_key;

        Scroll(1, 0, nScreenRows, nScreenColumns, -1, nNormal);
        DisplayLine(1, LFile.lViewPtr);
        continue;

      case CUR_DOWN:
        if (ListMoveLine(1) == 0)
          goto bad_key;

        Scroll(1, 0, nScreenRows, nScreenColumns, 1, nNormal);

        // display last line
        lTemp = LFile.lViewPtr;
        lRow = (nScreenRows - 1);
        lTemp += MoveViewPtr(lTemp,&lRow);
        if (lRow == (nScreenRows - 1))
          DisplayLine(nScreenRows, lTemp);

        continue;

      case HOME:
        ListHome();
        break;

      case END:
        // goto the last row
        list_wait(LIST_WAIT);
        while (ListMoveLine(1) != 0);

      case PgUp:
         // already at TOF?
         if (LFile.lViewPtr == 0L)
           goto bad_key;

         i = 1;
         while ((i < nScreenRows) && (ListMoveLine(-1) != 0))
           i++;

         break;

      case PgDn:
      case SPACE:
        if (ListMoveLine(nScreenRows - 1) == 0)
          goto bad_key;

        break;

      case F1:        // help
        _help("LIST", NULL);
        continue;

      case LIST_INFO_CHAR:
        {
        ULONG size=1024L;
        PCH fptr;
        INT fFSType;
        CHAR szBuf[10];

        // disable ^C / ^BREAK handling
        HoldSignals();

        if ((uListFlags & LIST_STDIN) == 0)
          {
          if (find_file(FIND_FIRST, LFile.szName, 0x2007, &dir, NULL) == NULL)
            {
            honk();
            continue;
            }
          }

        // display info on the current file
        i = ((uListFlags & LIST_STDIN) ? 5 : 10);
        fFSType = ifs_type(LFile.szName);
        if (fFSType != FAT)
          i = 11;

        wn = wOpen(2, 3, i, 75, nInverse, gpInternalName, NULL);

        wn->attrib = nNormal;
        wClear();
        i = 0;

        if (uListFlags & LIST_STDIN)
          wWriteStrAtt(0, 1, nNormal, LIST_INFO_PIPE);
        else
          {
          szDescription[0] = '\0';
          process_descriptions(LFile.szName, szDescription, DESCRIPTION_READ);

          fptr = AllocMem(size);

          strcpy(szLine, FormatDate(dir.fdLW.fdateLastWrite.month,
                         dir.fdLW.fdateLastWrite.day,
                         dir.fdLW.fdateLastWrite.year + 80));
          strcpy(szBuf, FormatDate(dir.fdLA.fdateLastAccess.month,
                                   dir.fdLA.fdateLastAccess.day,
                                   dir.fdLA.fdateLastAccess.year + 80));
          sprintf(fptr, ((fFSType != FAT) ? LIST_INFO_HPFS : LIST_INFO_FAT),
                  LFile.szName, szDescription, dir.size, szLine,
                  dir.ftLW.ftimeLastWrite.hours,
                  gaCountryInfo.szTimeSeparator[0],
                  dir.ftLW.ftimeLastWrite.minutes, szBuf,
                  dir.ftLA.ftimeLastAccess.hours,
                  gaCountryInfo.szTimeSeparator[0],
                  dir.ftLA.ftimeLastAccess.minutes,
                  FormatDate(dir.fdC.fdateCreation.month,
                             dir.fdC.fdateCreation.day,
                             dir.fdC.fdateCreation.year + 80),
                  dir.ftC.ftimeCreation.hours,
                  gaCountryInfo.szTimeSeparator[0],
                  dir.ftC.ftimeCreation.minutes);

          // print the text
          for (; (*fptr != '\0'); i++)
            {
            sscanf(fptr, "%[^\n]%*c%n", szHeader, &size);
            wWriteStrAtt(i, 1, nNormal, szHeader);
            fptr += size;
            }

          FreeMem(fptr);
          }  // else

        wWriteListStr(i+1,1,wn,PAUSE_PROMPT);
        GetKeystroke(EDIT_NO_ECHO | EDIT_BIOS_KEY);

        wRemove(wn);

        // enable ^C / ^BREAK handling
        EnableSignals();
        continue;
        }

      case LIST_GOTO_CHAR:
        // goto the specified line / hex offset

        // disable ^C / ^BREAK handling
        HoldSignals();

        wn = wOpen(2, 5, 4, strlen(LIST_GOTO_OFFSET) + 20, nInverse,
                   LIST_GOTO_TITLE, NULL);
        wn->attrib = nNormal;
        wClear();

        wWriteListStr(0, 1, wn,
                   ((uListFlags & LIST_HEX) ? LIST_GOTO_OFFSET : LIST_GOTO));
        i = egets(szLine, 10, (EDIT_DIALOG | EDIT_BIOS_KEY | EDIT_NO_CRLF));
        wRemove(wn);

        // enable ^C / ^BREAK handling
        EnableSignals();

        if (i == 0)
          break;

        list_wait(LIST_WAIT);

        // if in hex mode, jump to offset
        if (uListFlags & LIST_HEX)
          {
          strupr(szLine);
          sscanf(szLine, "%lx", &lRow);
          lRow = lRow / 0x10;
          }
        else
          {
          if (sscanf(szLine, FMT_LONG, &lRow) == 0)
            continue;
          }

        lRow -= gpIniptr->ListRowStart;
        if (lRow >= 0)
          {
          LFile.lViewPtr = MoveViewPtr(0L, &lRow);
          LFile.lCurrentLine = lRow;
          }
        else
          {
          LFile.lViewPtr += MoveViewPtr(LFile.lViewPtr, &lRow);
          LFile.lCurrentLine += lRow;
          }

        break;

      case LIST_HIBIT_CHAR:
        // toggle high bit filter
        uListFlags ^= LIST_HIBIT;
        break;

      case LIST_WRAP_CHAR:
        // toggle line wrap
        uListFlags ^= LIST_WRAP;
        nRightMargin = ((uListFlags & LIST_WRAP) ? GetScrCols() : 512);

        // recalculate current line
        list_wait(LIST_WAIT);

        // get start of line
        LFile.nListHorizOffset = 0;

        // line number probably changed, so recompute everything
        LFile.lCurrentLine = ComputeLines(0L, LFile.lViewPtr);
        LFile.lViewPtr = MoveViewPtr(0L, &(LFile.lCurrentLine));
        break;

      case LIST_HEX_CHAR:
        // toggle hex display
        uListFlags ^= LIST_HEX;

        // if hex, reset to previous 16-byte
        //   boundary
        if (uListFlags & LIST_HEX)
          LFile.lViewPtr -= (LFile.lViewPtr % 16);
        else
          {
          // if not hex, reset to start of line
          ListMoveLine(1);
          ListMoveLine(-1);
          }

        // recalculate current line
        list_wait(LIST_WAIT);
        LFile.lCurrentLine = ComputeLines(0L, LFile.lViewPtr);
        break;

      case LIST_FIND_CHAR:
      case LIST_FIND_CHAR_REVERSE:
        // find first matching string
        bListSkipLine = 0;

      case LIST_FIND_NEXT_CHAR:
      case LIST_FIND_NEXT_CHAR_REVERSE:
        // find next matching string

        if ((c == LIST_FIND_CHAR) || (c == LIST_FIND_CHAR_REVERSE))
          {
          // disable ^C / ^BREAK handling
          HoldSignals();

          fSearchFlags = 0;
          wn = wOpen(2, 1, 4, 75, nInverse,
                     ((c == LIST_FIND_NEXT_CHAR_REVERSE)
                       ? LIST_FIND_TITLE_REVERSE : LIST_FIND_TITLE), NULL);
          wn->attrib = nNormal;
          wClear();

          if (uListFlags & LIST_HEX)
            {
            wWriteListStr(0, 1, wn, LIST_FIND_HEX);
            if (GetKeystroke(EDIT_ECHO | EDIT_BIOS_KEY | EDIT_UC_SHIFT) == YES_CHAR)
              fSearchFlags |= FFIND_HEX_SEARCH;

            wClear();
            }

          wWriteListStr(0, 1, wn, LIST_FIND);
          egets(szListFindWhat, 64,
                (EDIT_DIALOG | EDIT_BIOS_KEY | EDIT_NO_CRLF));
          wRemove(wn);

          // enable ^C / ^BREAK handling
          EnableSignals();
          }
        else
          {
FindNext:
          // a "Next" has to be from current position
          fSearchFlags &= ~FFIND_TOPSEARCH;
          }

        if (szListFindWhat[0] == '\0')
          continue;

        sprintf(szDescription, LIST_FIND_WAIT, szListFindWhat);
                list_wait(szDescription);

        // save start position
        lTemp = LFile.lViewPtr;
        lRow = LFile.lCurrentLine;

        if ((c == LIST_FIND_CHAR_REVERSE) || (c == LIST_FIND_NEXT_CHAR_REVERSE))
          {
          // start on the previous line
          fSearchFlags |= FFIND_REVERSE_SEARCH;
          }
        else
          {
          fSearchFlags &= ~FFIND_REVERSE_SEARCH;

          // skip the first line (except on /T"xxx")
          if (bListSkipLine)
            ListMoveLine(1);
          }

        bListSkipLine = 1;
        if (SearchFile(fSearchFlags) != 1)
          {
          ListSetCurrent(lTemp);
          LFile.lViewPtr = lTemp;
          LFile.lCurrentLine = lRow;
          }
        else
          LFile.fDisplaySearch = ((fSearchFlags & FFIND_CHECK_CASE) ? 2 : 1);

        break;

      case LIST_PRINT_CHAR:
        // print the file
        ListPrintFile();
        continue;

      case LIST_CONTINUE_CHAR:
      case CTL_PgDn:
        return 0;

      case LIST_PREVIOUS_CHAR:
      case CTL_PgUp:
        // previous file
        if (nCurrent > 0)
          {
          nCurrent--;
          return 0;
          }

      default:
bad_key:
        honk();
        continue;
      }

    // rewrite the display
    ListUpdateScreen();
    }

  return 0;
  }
/************************************************************************/
// display a "WAIT" message
static void list_wait(char *pszPrompt)
  {
  // clear the header line
  clear_header();

  // display "WAIT" in blinking chars
  WriteStrAtt(0, ((nScreenColumns / 2) - (strlen(pszPrompt) / 2)),
              (nInverse | 0x80), pszPrompt);
  }
/************************************************************************/
// clear the LIST header line to inverse blanks
static void clear_header(void)
  {
  dirty_header = 1;
  Scroll(0, 0, 0, nScreenColumns, 0, nInverse);
  }
/************************************************************************/
// redraw the screen (except for the status line)
static void ListUpdateScreen(void)
  {
  int nRow;
  long lTemp = LFile.lViewPtr;

  Scroll(1, 0, nScreenRows, nScreenColumns, 0, nNormal);

  // if no name yet, we're still initializing!
  if (LFile.szName[0] == '\0')
    return;

  for (nRow = 1; (nRow <= nScreenRows); nRow++)
     lTemp += DisplayLine(nRow, lTemp);

  LFile.fDisplaySearch = 0;
  }
/************************************************************************/
// print the file on the default printer
void ListPrintFile(void)
  {
  int i, n;
  int c, nRows, nBytesPrinted;
  long lTemp;
  char szBuffer[512];
  POPWINDOWPTR wn;

  // disable ^C / ^BREAK handling
  HoldSignals();

  wn = wOpen(2, 1, 4, strlen(LIST_PRINTING) + 8, nInverse, LIST_PRINT_TITLE,
             NULL);
  wn->attrib = nNormal;
  wClear();
  sprintf(szBuffer, LIST_QUERY_PRINT, LIST_PRINT_FILE_CHAR,
          LIST_PRINT_PAGE_CHAR);
  wWriteListStr(0, 1, wn, szBuffer);

  c = GetKeystroke(EDIT_ECHO | EDIT_UC_SHIFT | EDIT_BIOS_KEY);
  if ((c == LIST_PRINT_FILE_CHAR) || (c == LIST_PRINT_PAGE_CHAR))
    {
    // save start position
    lTemp = LFile.lViewPtr;

    // display "Printing ..."
    wWriteListStr(0, 1, wn, LIST_PRINTING);

    gnGFH = _sopen(((gpIniptr->Printer != INI_EMPTYSTR)
                    ? gpIniptr->StrData + gpIniptr->Printer : "LPT1"),
                    (O_TEXT | O_WRONLY | O_CREAT), SH_DENYNO,
                    S_IWRITE | S_IREAD);
    if (gnGFH >= 0)
      {
      // reset to beginning of file
      if (c == LIST_PRINT_FILE_CHAR)
        ListSetCurrent(0L);
      else
        {
        nRows = GetScrRows();
        ListSetCurrent(LFile.lViewPtr);
        }

      // print the header (filename, date & time)
      qprintf(gnGFH, "%s   %s  %s\n\n", LFile.szName, gdate(0),
              gtime(gaCountryInfo.fsTimeFmt));

      do
        {
        // abort printing if a key is hit
        if ((_kbhit() != 0)
            && (GetKeystroke(EDIT_NO_ECHO | EDIT_BIOS_KEY) == ESC))
          break;

        i = FormatLine(szBuffer, LFile.lViewPtr, &nBytesPrinted, TRUE);
        LFile.lViewPtr += 16;

        // replace 0-31 with "."
        if (uListFlags & LIST_HEX)
          {
          for (n = 0; (n < i); n++)
            {
            if (szBuffer[n] < 32)
              szBuffer[n] = '.';
            }
          }

        if ((c == LIST_PRINT_PAGE_CHAR) && (nRows-- <= 0))
          break;
        } while ((nBytesPrinted > 0)
                 && (qprintf(gnGFH, "%.*s\n", i, szBuffer) > 0));

      // print a formfeed
      qputc(gnGFH, '\014');
      gnGFH = _close(gnGFH);

      // restore start position
      LFile.lViewPtr = lTemp;
      ListSetCurrent(LFile.lViewPtr);
      }
    else
      honk();
    }

  wRemove(wn);

  // enable ^C / ^BREAK handling
  EnableSignals();
  }
/************************************************************************/
// write a string to a popup window & position cursor at end
static void wWriteListStr(int nRow, int col, POPWINDOWPTR wn, char *str)
  {
  wWriteStrAtt(nRow, col, wn->attrib, str);
  wSetCurPos(nRow, col + strlen(str));
  }


