#define INCL_BASE
#define INCL_WIN
#define INCL_GPI
#include <os2.h>
#include <bsedos.h>

// Prototype for DosQProcStatus()
//   imp DOSQPROCSTAT DOSCALL1.154

#if defined __cplusplus
extern "C"
{
#endif
USHORT _Far16 _Pascal DosQProcStatus(PVOID pBuf, USHORT cbBuf);

#if defined __cplusplus
}
#endif

/* Size of the buffer into which process status info will be read */
#define BUFSIZE  64000

// structures needed to interpret the process information returned by
// DosQProcStatus(), plus some other things that might be generally useful

typedef struct _node
  {
  USHORT code;      // type of node
  USHORT next;      // offset within returned data of the
                    // next node. This is actually the
                    // offset from the SEGMENT containing
                    // the buffer, not from the buffer
                    // itself.  Ugly, huh?
  } node;

// These are the possible values of node.code:

#define PNODE    0   // A process node: points to other data
#define TNODE    1   // A thread node: thread info
#define MNODE    2   // A module node: module info
#define SEMNODE  3   // A system semaphone node
#define SSNODE   4   // A shared segment node

// There is one special value for next:

#define EOC        0xffff             // End of chain

// Process node:  contains pid and ppid, and points to associated info

typedef struct _procnode
  {
  USHORT code;        // node type: 0
  USHORT next;        // offset of next node
  USHORT pid;         // process id
  USHORT ppid;        // process id of parent
  USHORT sid;         // session id
  USHORT hmod;        // module handle of executable code
  USHORT modinfo;     // offset of first mnode
  USHORT mnodecnt;    // number of mnodes associated
  USHORT ssinfo;      // offset of first ssnode
  USHORT ssnodecnt;   // number of ssnodes associated
  USHORT seginfo;     // offset of first segnode
  USHORT segnodecnt;  // number of segnodes associated
  } procnode;

// There is no explicit pointer or counter for thread nodes.  The first one
// follows immediately the process node.

typedef struct _threadnode
  {
  USHORT code;        // node type: 1
  USHORT next;        // offset of next thread node
  USHORT hmod;        // module handle of executable code
  USHORT pid;         // process id
  USHORT tid;         // thread id
  USHORT priority;    // thread priority
  USHORT state;       // thread state
  ULONG blockid;      // block id
  } threadnode;

// These are the possible thread states

#define TSTART          0
#define TZOMBIE         1
#define TREADY          2
#define TBLOCKED        3
#define TFROZEN         4
#define TCRITSEC        5
#define TBACKGROUND     6
#define TBOOST          7

// Structure of a module node
typedef struct _modulenode
  {
  USHORT code;        // node type: 2
  USHORT next;
  USHORT hmod;
  USHORT depcount;    // number of dependent modules
  USHORT deps;        // offset of first dependent module node
  USHORT name;        // offset of 0-terminated module name
  USHORT deplist;     // dependency list ????
  } modulenode;

// Structure of a semaphore node
typedef struct _semaphorenode
  {
  USHORT code;      // note type:  3
  USHORT next;
  USHORT flag;
  BYTE refs;        // number of references
  BYTE reqs;        // number of requests outstanding
  USHORT index;     // ?
  BYTE name;        // name of semaphore in ascii
  } semaphorenode;

// Structure of a shared segment node
typedef struct _ssegnode
  {
  USHORT code;     // 4
  USHORT next;
  USHORT handle;   // handle of the shared segment
  SEL sel;         // selector of the shared segment
  USHORT refs;     // number of references to this ss
  CHAR name;       // segment name
  } ssegnode;

// Layout of the Global Record (next 12 bytes)
typedef struct _qsGrec
  {
  ULONG cThrds;   // active threads in system at time of call
  ULONG Reserved1;
  ULONG Reserved2;
  } qsGrec;

/* each Process record is followed by
       - a list of 16-bit semephore indices (c16Sem of them).
       - a list of dll mte handles (cLib of them).
       - a list of shared memory handles (cShrMem of them).
       - 1 or more thread records (cTCB of them).
*/

// Layout of the thread record
#define THREAD_RECORD_TYPE  256
typedef struct _qsTrec
  {
  ULONG RecType;     // THREAD_RECORD_TYPE
  USHORT tid;        // relative thread id
  USHORT slot;       // unique thread id
  ULONG sleepid;     // sleep id if thread is sleeping
  ULONG priority;    // thread priority
  ULONG systime;     // thread system time
  ULONG usertime;    // thread user time
  BYTE state;        // thread state
  BYTE PAD1;
  USHORT PAD2;
  } qsTrec;

// Layout of the Process record (idendified by record type 1)
#define PROCESS_RECORD_TYPE      1
typedef struct _qsPrec
  {
  ULONG RecType;              // PROCESS_RECORD_TYPE
  qsTrec *pThrdRec;           // First thread record
  USHORT pid;                 // Process id
  USHORT ppid;                // Parent pid
  ULONG type;                 // Process type
  ULONG stat;                 // Process status
  ULONG sgid;                 // Screen group id
  USHORT hMte;                // Module handle
  USHORT cTCB;                // number of threads
  ULONG Reserved1;
  PVOID Reserved2;
  USHORT c16Sem;              // number of 16-bit sems
  USHORT cLib;                // number of dlls
  USHORT cShrMem;             // number of shared mem handles
  USHORT Reserved3;
  PUSHORT p16SemInd;          // ptr to 16-bit sem indices
  PUSHORT pLibRec;            // ptr to dll mte handles
  PUSHORT pShrMemRec;         // ptr to shared mem handles
  PVOID Reserved4;
  } qsPrec;

// Layout of the DLL Module Record
typedef struct _qsLrec
  {
//  struct LibRec *pNextRec;    // next DLL record
  PVOID pNextRec;
  USHORT hmte;                // handle for this module
  USHORT Reserved1;
  ULONG cImpMod;              // number of imported modules
  ULONG Reserved2;
  PVOID Reserved3;
  PBYTE pName;                // module path and name
  } qsLrec;

// Layout of the shared memory segment record
typedef struct _qsMrec
  {
  ULONG MemNextRec;
  USHORT hmem;           // handle
  USHORT sel;            // shared mem selector
  USHORT refcnt;         // reference count
  BYTE MemName;          // first byte of name
  } qsMrec;


// Here are some of the thread states:
#define TS_READY   1
#define TS_BLOCKED 2
#define TS_RUNNING 5

/* The DLL record is followed by a list of imported module handles (cImpMod
   of them); and by the pathname (pointed to by pName), which is an
   ASCIIZ string.
*/

// Layout of the 16-bit semaphore record
typedef struct _qsS16Head
  {
  ULONG SRecType;
  ULONG SpNextRec;
  ULONG S32SemRec;
  ULONG S16TblOff;
  ULONG pSem16Rec;
  } qsS16Head;

typedef struct _SysSemTbl
  {
  ULONG s_SysSemOwner;   // owning thread
  BYTE s_SysSemFlag;     // bit flags
  BYTE s_SysSemRefCnt;   // number of references
  BYTE s_SysSemProcCnt;  // requests by owner
  BYTE Reserved1;
  ULONG Reserved2;
  ULONG Reserved3;
  } SysSemTbl;

typedef struct _qsS16rec
  {
  ULONG NextRec;
  SysSemTbl SysSemData;
  BYTE SemName;    // first byte of asciiz name
  } qsS16rec;

// Layout of the pointer record (first 24 bytes of the structure)
typedef struct _qsPtrrec
  {
  qsGrec *pGlobalRec;       // 'global' record
  qsPrec *pProcRec;         // first process record
  qsS16Head *p16SemRec;     // 1st 16-bit sem record
  PVOID Reserved1;          // ***not in docs***
  qsMrec *pShrMemRec;       // 1st shared mem record
  qsLrec *pLibRec;          // 1st module (mte) rec
  } qsPtrrec;

