// SELECT.C - SELECT commands for 4xxx / TCMD family
//   Copyright (c) 1993 - 1997 Rex C. Conn  All rights reserved
// (c) 2003 Doug Lindquist

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <io.h>

#include "4all.h"
#include "product.h"

// SELECT arguments
typedef struct
  {
  UINT marked;
  INT dirline;
  INT dirfirst;
  INT dirlast;
  UINT normal;
  UINT inverse;
  UINT entries;
  UINT length;
  UINT uMarkedFiles;
  INT nLeftMargin;
  LONG lTotalSize;
  CHAR *arg;
  CHAR *cmd_start;
  CHAR *cmd_tail;
  CHAR *pFileVar;
  CHAR open_paren;
  CHAR close_paren;
  CHAR szPath[MAXPATH+1];
  DIR_ENTRY *sfiles;
  RANGES aSelRanges;
  } SELECT_ARGS;

static int InitSelect(SELECT_ARGS *, char *, int *);
static int ProcessSelect(SELECT_ARGS *, char *);
static VOID ShowSelectHeader(SELECT_ARGS *);
static int getmark(SELECT_ARGS *sel);
static VOID dirprint(SELECT_ARGS *, int, UINT, UINT);

/************************************************************************/
/************************************************************************/
// select the files for a command
INT select_cmd(int argc, char **argv)
  {
  PSZ arg, selectline;
  INT rval=0, mode=0x10;
  SELECT_ARGS sel;

  selectline = *argv;

  // check for SELECT options
  arg = argv[1];
  if (arg == NULL)
    return (usage(SELECT_USAGE));

  rval = InitSelect(&sel, arg, &mode);
  if (rval != 0)
    return rval;

  if (*arg == '\0')
    return (usage(SELECT_USAGE));

  sel.pFileVar = strpbrk(arg, "([");
  if (sel.pFileVar == NULL)
    return (usage(SELECT_USAGE));

  sel.cmd_tail = scan(sel.pFileVar, ((*sel.pFileVar == '(') ? ")" : "]"), "`[");
  if ((sel.cmd_tail == BADQUOTES) || (sel.cmd_tail[0] == '\0'))
    return (usage(SELECT_USAGE));

  sel.open_paren = *sel.pFileVar;
  sel.close_paren = sel.cmd_tail[0];

  // rub out the '(' & ')' (or [])
  *sel.pFileVar++ = '\0';
  *sel.cmd_tail++ = '\0';

  // save the command on the stack
  sel.cmd_start = _alloca(strlen(arg) + 1);
  strcpy(sel.cmd_start, arg);
  arg = _alloca(strlen(sel.cmd_tail) + 1);
  sel.cmd_tail = strcpy(arg, sel.cmd_tail);

  // do any variable expansion on the file spec
  if (var_expand(strcpy(selectline, sel.pFileVar), 1) != 0)
    return ERROR_EXIT;

  sel.pFileVar = _alloca(strlen(selectline) + 1);
  strcpy(sel.pFileVar, selectline);

  argc = 0;
  sel.arg = ntharg(sel.pFileVar, argc);
  while ((sel.arg != NULL) && (rval != CTRLC))
    {
    sel.sfiles = 0L;        // initialize pointer to file array
    sel.entries = 0;

    if (setjmp(cv.env) == -1)
      {
      dir_free(sel.sfiles);
      cls_cmd(1, NULL);
      rval = CTRLC;
      goto select_bye;
      }

    EnableSignals();
    copy_filename(selectline, sel.arg);

    // force \*.* onto a directory name
    if (is_dir(selectline))
      mkdirname(selectline, WILD_FILE);

    // save path part of name
    arg = path_part(selectline);
    if (arg == NULL)
      arg = NULLSTR;

    strcpy(sel.szPath, arg);

    if (mkfname(selectline, 0) == NULL)
      {
      rval = ERROR_EXIT;
      goto select_bye;
      }

    // check to see if it's an HPFS partition
    if (ifs_type(selectline) != 0)
      glDirFlags |= DIRFLAGS_HPFS;

    // search the directory for the specified file(s)
    if (SearchDirectory((mode | FIND_DATERANGE), selectline,
                        (DIR_ENTRY **)&(sel.sfiles), &(sel.entries),
                        &(sel.aSelRanges), 5) != 0)
      {
      rval = ERROR_EXIT;
      goto select_bye;
      }

    if (sel.entries > 0)
      {
      // clear the screen
      cls_cmd(1, NULL);

      // get default normal and inverse attributes
      if (gpIniptr->SelectColor != 0)
        Scroll(0, 0, GetScrRows(), GetScrCols()-1, 0, gpIniptr->SelectColor);

      // get default screen colors
      GetAtt(&(sel.normal), &(sel.inverse));
      if (gpIniptr->SelectStatusColor != 0)
        sel.inverse = gpIniptr->SelectStatusColor;

      // display SELECT header
      ShowSelectHeader(&sel);

      // get filename colors
      ColorizeDirectory(sel.sfiles, sel.entries, 0);

      sel.dirline = sel.dirfirst = 0;     // beginning of page

      // get the selection list
      if (getmark(&sel) == -1)
        sel.marked = 0;

      cls_cmd(1, NULL);

      if (sel.length >= MAXLINESIZ)
        {
        rval = error(ERROR_4DOS_COMMAND_TOO_LONG, NULL);
        sel.marked = 0;
        }

      // process the marked files in order
      rval = ProcessSelect(&sel, selectline);
      }  // if (sel.entries > 0)

    dir_free(sel.sfiles);
    argc++;
    sel.arg = ntharg(sel.pFileVar, argc);
    }  // while ((sel.arg != NULL) && (rval != CTRLC))

select_bye:
  EnableSignals();
  return rval;
  }
/************************************************************************/
static VOID ShowSelectHeader(SELECT_ARGS *sel)
  {
  // display SELECT header
  Scroll(0, 0, 0, GetScrCols()-1, 0, sel->inverse);
  SetCurPos(0, 0);
  color_printf(sel->inverse, SELECT_HEADER, gchVerticalBar, gchVerticalBar,
               gchVerticalBar, gchVerticalBar);

  // display the command & the file spec(s)
  Scroll(1, 0, 1, GetScrCols()-1, 0, sel->normal);
  SetCurPos(1, 0);
  color_printf(sel->normal, "%s%c%s%c%s", sel->cmd_start, sel->open_paren,
               sel->pFileVar, sel->close_paren, sel->cmd_tail);
  }
/************************************************************************/
// SELECT the files to mark
static int getmark(SELECT_ARGS *sel)
  {
  extern INT ListEntry(char *);
  UINT c, i, j, n;
  INT nSearch=0;
  CHAR szFilename[MAXFILENAME+1], szSearch[32];
  jmp_buf saved_env;

  // display first page
  dirprint(sel, 1, sel->dirline, sel->inverse);

  szSearch[0] = '\0';
  sel->marked = 0;
  for (; ;)
    {
    c = cvtkey(GetKeystroke(EDIT_NO_ECHO | EDIT_UC_SHIFT), MAP_GEN);
    switch (c)
      {
      case ESC:
        return -1;

      case CR:
      case LF:
        // allow a single selection using the RETURN key
        if (sel->marked == 0)
          sel->sfiles[sel->dirline].file_mark = ++(sel->marked);

        return sel->marked;

      case SELECT_LIST:
        // if HPFS names with whitespace, quote them
        if (glDirFlags & DIRFLAGS_HPFS)
          {
          sprintf(szFilename, "%s%Fs", sel->szPath,
                  sel->sfiles[sel->dirline].hpfs_name);
          AddQuotes(szFilename);
          }
        else
          sprintf(szFilename, "%s%Fs", sel->szPath,
                  sel->sfiles[sel->dirline].file_name);

        // save the old "env" (^C destination)
        memmove(saved_env, cv.env, sizeof(saved_env));
        ListEntry(szFilename);

        // restore the old "env"
        memmove(cv.env, saved_env, sizeof(saved_env));

        // redraw SELECT screen
        ShowSelectHeader(sel);
        goto redraw_page;

      case '+':       // mark a file
      case SPACE:
        if (sel->sfiles[sel->dirline].file_mark == 0)
          sel->sfiles[sel->dirline].file_mark = ++(sel->marked);
        else

      case '-':       // unmark a file
          sel->sfiles[sel->dirline].file_mark = 0;

      case CUR_UP:
      case CUR_DOWN:
        // switch current line to normal
        dirprint(sel, -1, sel->dirline, sel->normal);
        if (c != CUR_UP)
          {
          // move scroll bar down, beep if at end
          if (sel->dirline >= ((int)sel->entries - 1))
            {
            if (c == CUR_DOWN)
              honk();
            }
          else
            {
            if (++(sel->dirline) > sel->dirlast)
              {
              (sel->dirfirst)++;
              Scroll(2, 0, GetScrRows(), GetScrCols()-1, 1, sel->normal);
              }
            }
          }  // if (c != CUR_UP)
        else
          {
          // move scroll bar up
          if (sel->dirline == 0)
            honk();     // already at top
          else
            {
            if (--(sel->dirline) < sel->dirfirst)
              {
              (sel->dirfirst)--;
              Scroll(2, 0, GetScrRows(), GetScrCols()-1, -1, sel->normal);
              }
            }
          }  // else

        // new line in inverse video
        dirprint(sel, -1, sel->dirline, sel->inverse);
        break;

      case HOME:
        sel->nLeftMargin = sel->dirfirst = sel->dirline = 0;

      case END:
      case PgUp:
      case PgDn:
        i = GetScrRows() - 1;

        // change the directory page
        if (c == END)
          {
          sel->dirline = sel->entries - 1;
          sel->dirfirst = ((sel->dirline + 1) - i);
          if (sel->dirfirst < 0)
            sel->dirfirst = 0;
          }
        else
          {
          if (c == PgDn)
            {
            if ((int)(sel->dirline += i) >= (int)(sel->entries))
              sel->dirline = sel->entries - 1;

            if ((int)(sel->dirfirst + i) < (int)(sel->entries))
              sel->dirfirst += i;
            }
          else
            {
            if (c == PgUp)
              {
              if ((sel->dirfirst -= i) < 0)
                sel->dirfirst = 0;

              if ((sel->dirline -= i) < 0)
                sel->dirline = 0;
              }
            }
          }

      case '*':       // reverse all current marks (no dirs)
      case '/':       // unmark everything
        for (i = 0; ((c < 0x80) && (i < sel->entries)); i++)
          {
          if (c == '*')
            {
            if (sel->sfiles[i].file_mark == 0)
              {
              if ((sel->sfiles[i].attribute & _A_SUBDIR) == 0)
                sel->sfiles[i].file_mark = ++sel->marked;
              }
            else
              sel->sfiles[i].file_mark = 0;
            }
          else
            {
            if (c == '/')
              sel->sfiles[i].file_mark = sel->marked = 0;
            }
          }

      case CUR_LEFT:
      case CUR_RIGHT:
        if (c == CUR_LEFT)
          {
          sel->nLeftMargin -= 8;
          if (sel->nLeftMargin < 0)
            sel->nLeftMargin = 0;
          }
        else
          {
          if (c == CUR_RIGHT)
            {
            if (sel->nLeftMargin < 512)
              sel->nLeftMargin += 8;
            }
          }

redraw_page:
        // draw a new page
        dirprint(sel,1,sel->dirline,sel->normal);
        dirprint(sel,-1,sel->dirline,sel->inverse);
        break;

      case F1:    // help
        _help(gpInternalName, NULL);
        break;

      default:
        // try to match a string
        if (c <= 0xFF)
          {
          // search for entry beginning with szSearch
ResetSearch:
          if (nSearch >= 32)
            nSearch = 0;

          szSearch[nSearch++] = c;
          szSearch[nSearch] = '\0';
          i = sel->dirline;
          for (j = 0; (i < sel->entries); i++)
            {
StartLoop:
            if (((glDirFlags & DIRFLAGS_HPFS) == 0)
                || (glDirFlags & DIRFLAGS_HPFS_TO_FAT))
              n = strnicmp(szSearch, sel->sfiles[i].file_name,
                           strlen(szSearch));
            else
              n = strnicmp(szSearch, sel->sfiles[i].hpfs_name,
                           strlen(szSearch));

            if (n == 0)
              {
              // switch current line to normal
              dirprint(sel, -1, sel->dirline, sel->normal);
              sel->dirline = i;

              // draw page
              if ((sel->dirline < sel->dirfirst)
                  || (sel->dirline > sel->dirlast))
                {
                sel->dirfirst = sel->dirline;
                dirprint(sel, 1, sel->dirline, sel->normal);
                dirprint(sel, -1, sel->dirline, sel->inverse);
                }
              else
                dirprint(sel, -1, sel->dirline, sel->inverse);

              goto NoHonk;
              }

            if (i == (sel->entries - 1))
              {
              if (j == 0)
                {
                // try again from the beginning
                j = 1;
                i = 0;
                goto StartLoop;
                }
              else
                {
                if (j == 1)
                  {
                  if (nSearch == 1)
                    {
                    nSearch = 0;
                    break;
                    }

                  nSearch = 0;
                  goto ResetSearch;
                  }
                }
              }
            }  // for (j = 0; (i < sel->entries); i++)
          }  // if (c <= 0xFF)

      // invalid input
      honk();

NoHonk:
      break;
      }  // switch (c)
    }  // for (; ;)
  }
/************************************************************************/
// print directory for SELECT
//  page = directory page
//  line = line to print
//  attr = attribute to use for printing
static VOID dirprint(SELECT_ARGS *sel, int page, UINT line, UINT attr)
  {
  INT i, n, columns, rows;
  UINT current_row, save_row, last, color, nOffset;
  CHAR szBuffer[256];
  PCH fptr;

  sel->uMarkedFiles = 0;

  current_row = save_row = 2;     // set start row
  rows = GetScrRows();
  last = rows - 2;

  columns = GetScrCols();

  if (page > 0)
    {
    // print entire page
    // clear screen
    Scroll(2,0,rows,columns-1,0,sel->normal);
    line = sel->dirfirst;
    }
  else              // just print 1 line
    current_row += (line - sel->dirfirst);

  sel->dirlast = (sel->dirfirst + last);
  if (sel->dirlast >= (int)sel->entries)
    sel->dirlast = sel->entries - 1;

  // print page number
  i = ((sel->dirline + 1) / (last + 1));
  if ((sel->dirline + 1) % (last + 1))
    i++;

  n = (sel->entries / rows);
  if (sel->entries % rows)
    n++;

  sprintf(szBuffer, SELECT_PAGE_COUNT, i, n);
  WriteStrAtt(0, columns - (strlen(SELECT_PAGE_COUNT) - 1), sel->inverse,
              szBuffer);

  // initialize the SELECT line length
  sel->length = strlen(sel->cmd_start) + strlen(sel->cmd_tail);

  if (sel->open_paren == '(')
    sel->length += strlen(sel->szPath) + 13;

  sel->lTotalSize = 0L;
  for (last = 0; (last < sel->entries); last++)
    {
    if (sel->sfiles[last].file_mark > 0)
      {
      sel->uMarkedFiles++;
      if (sel->open_paren == '[')
        {
        // HPFS partitions have filename in "hpfs_name"
        fptr = ((glDirFlags & DIRFLAGS_HPFS) ? sel->sfiles[last].hpfs_name
                                             : sel->sfiles[last].file_name);
        i = strlen(fptr);
        sel->length += strlen(sel->szPath) + i + 1;
        }

      sel->lTotalSize += (LONG)((sel->sfiles[last].file_size + 1023L) / 1024L);
      }  // if (sel->sfiles[last].file_mark > 0)
    }  // for (last = 0; (last < sel->entries); last++)

  // display the line length
  sprintf(szBuffer, "%4u", sel->length);
  WriteStrAtt(0, 0, sel->inverse, szBuffer);

  // display the number of files marked & the size (in Kb)
  sprintf(szBuffer, MARKED_FILES, sel->uMarkedFiles, sel->lTotalSize);
  WriteStrAtt(1, columns-(strlen(MARKED_FILES)+3), sel->normal, szBuffer);

  // last row to print
  last = ((page > 0) ? sel->dirlast : line);

  for (nOffset = 0; (line <= last); line++, current_row++)
    {
    // check for colorization
    // get background color if none specified
    if ((sel->sfiles[line].color & 0x70) == 0)
      sel->sfiles[line].color |= (sel->normal & 0x70);

    color = (((attr == sel->normal) && (sel->sfiles[line].color != -1))
             ? sel->sfiles[line].color : attr);

    if (((glDirFlags & DIRFLAGS_HPFS) == 0)
        || (glDirFlags & DIRFLAGS_HPFS_TO_FAT))
      {
      // display the FAT filename
      nOffset = sprintf(szBuffer, "%c%-12Fs",
                        ((sel->sfiles[line].file_mark > 0)
                         ? gchRightArrow : ' '),
                        sel->sfiles[line].file_name);
      }
    else
      {
      // display the HPFS filename (max of first 45 chars
      //   displayable on an 80-columns display)
      nOffset = sprintf(szBuffer, "%c%-*.*Fs",
                        ((sel->sfiles[line].file_mark > 0)
                         ? gchRightArrow : ' '),
                        (columns-35), (columns-35),
                        sel->sfiles[line].hpfs_name);
      }

    if (((glDirFlags & DIRFLAGS_HPFS) == 0)
        || (glDirFlags & DIRFLAGS_HPFS_TO_FAT))
      {
      nOffset += sprintf(szBuffer+nOffset,
                         (sel->sfiles[line].attribute & _A_SUBDIR)
                          ? DIR_LABEL : "%9lu", sel->sfiles[line].file_size);
      nOffset += sprintf(szBuffer+nOffset,
                         ((gaCountryInfo.fsTimeFmt == 0)
                          ? "  %s %2u%c%02u%c" : "  %s  %2u%c%02u"),
                          FormatDate(sel->sfiles[line].fd.file_date.months,
                                     sel->sfiles[line].fd.file_date.days,
                                     sel->sfiles[line].fd.file_date.years + 80),
                          sel->sfiles[line].ft.file_time.hours,
                          gaCountryInfo.szTimeSeparator[0],
                          sel->sfiles[line].ft.file_time.minutes,
                          sel->sfiles[line].ampm);
      {
      n = strlen(sel->sfiles[line].file_id);
      i = ((n > sel->nLeftMargin) ? sel->nLeftMargin : n);
      if ((n - i) > (columns - 40))
        sprintf(szBuffer+nOffset," %.*Fs%c",(columns-41),
                sel->sfiles[line].file_id+i, gchRightArrow);
      else
        sprintf(szBuffer+nOffset," %Fs", sel->sfiles[line].file_id+i);
      }

      WriteStrAtt(current_row, 0, color, szBuffer);
      SetLineColor(current_row, 39, (columns - 39), color);
      }  // if (((glDirFlags & DIRFLAGS_HPFS) == 0)...
    else
      {
      nOffset += sprintf(szBuffer+nOffset,
                         (sel->sfiles[line].attribute & _A_SUBDIR)
                          ? HPFS_DIR_LABEL : "%15Lu",
                         sel->sfiles[line].file_size);

      sprintf(szBuffer+nOffset, "  %s  %2u%c%02u%c",
              FormatDate(sel->sfiles[line].fd.file_date.months,
                         sel->sfiles[line].fd.file_date.days,
                         sel->sfiles[line].fd.file_date.years + 80),
              sel->sfiles[line].ft.file_time.hours,
              gaCountryInfo.szTimeSeparator[0],
              sel->sfiles[line].ft.file_time.minutes,
              sel->sfiles[line].ampm);
      WriteStrAtt(current_row, 0, color, szBuffer);
      }

    if (attr == sel->inverse)
      save_row = current_row;

    attr = sel->normal;
    }  // for (nOffset = 0; (line <= last); line++, current_row++)

  // move the cursor to the currently selected row (this is
  //   for blind users who need the cursor to move)
  SetCurPos(save_row, 0);
  }
/************************************************************************/
/************************************************************************/
// check for SELECT options
static int InitSelect(SELECT_ARGS *sel, char *start_line, int *mode)
  {
  int i;
  char *arg;

  // clear everything to 0
  memset((char *)sel, '\0', sizeof(SELECT_ARGS));

  // reset the DIR flags
  init_dir();

  // initialize date/time/size ranges
  if (GetRange(start_line, &(sel->aSelRanges), 1) != 0)
    return ERROR_EXIT;

  i = 0;
  arg = ntharg(start_line, i);
  while ((arg != NULL) && (*arg == gpIniptr->SwChr))
    {
    // point past switch character
    arg++;
    while (*arg != '\0')
      {
      switch (_ctoupper(*arg++))
        {
        case 'A':
          // retrieve based on specified atts
          arg = GetSearchAttributes(arg);

          // default to everything
          *mode = 0x417;
          break;

        case 'D':       // don't colorize
          glDirFlags |= DIRFLAGS_NO_COLOR;
          break;

        case 'E':       // display filenames in upper case
          glDirFlags |= DIRFLAGS_UPPER_CASE;
          break;

        case 'H':       // hide "." and ".."
          glDirFlags |= DIRFLAGS_NO_DOTS;
          break;

        case 'I':       // find matching descriptions
          if (*arg == '"')
            sscanf(++arg, "%511[^\"]", gszFindDesc);
          else
            strcpy(gszFindDesc, arg);

          arg = NULLSTR;
          break;

        case 'J':       // DOS justify filenames
          glDirFlags |= DIRFLAGS_JUSTIFY;
          break;

        case 'L':       // display filenames in lower case
          glDirFlags |= DIRFLAGS_LOWER_CASE;
          break;

        case 'O':       // dir sort order
          // kludges for DOS 5 format
          arg = dir_sort_order(arg);
          break;

        case 'T':       // display attributes or time field
          if (*arg)
            {
            if (*arg == ':')
              arg++;

            switch (_ctoupper(*arg))
              {
              case 'A':       // last access
                gnDirTimeField = 1;
                break;

              case 'C':       // creation
                gnDirTimeField = 2;
                break;

              case 'W':       // last write (default)
                gnDirTimeField = 0;
                break;

              default:
                goto SelectError;
              }

            arg = NULLSTR;
            }
          else
            glDirFlags |= DIRFLAGS_SHOW_ATTS;

          break;   // case 'T'

        case 'Z':
          glDirFlags |= DIRFLAGS_HPFS_TO_FAT;
          break;

        default:  // invalid option
SelectError:
          error(ERROR_INVALID_PARAMETER, arg-1);
          return (usage(SELECT_USAGE));
        }
      }  // while (*arg != '\0')

    i++;
    arg = ntharg(start_line, i);
    }  // while (arg != NULL) && (*arg == gpIniptr->SwChr)

  // skip the switch statements
  strcpy(start_line, ((gpNthptr == NULL) ? NULLSTR : gpNthptr));
  return 0;
  }
/************************************************************************/
// process the SELECT'd filenames
static int ProcessSelect(SELECT_ARGS *sel, char *szCmdline)
  {
  UINT i, n;
  int rval = 0;
  long lSaveDirFlags;
  char *szFormat;

  // process the marked files in order
  *szCmdline = '\0';
  n = 0;
  i = 1;
  while ((rval != CTRLC) && (i <= sel->marked))
    {
    if (setjmp(cv.env) == -1)
      {
      cls_cmd(1, NULL);
      rval = CTRLC;
      break;
      }

    if (sel->sfiles[n].file_mark == i)
      {
      if (sel->open_paren == '(')
        {
        // substitute the filename & run the
        //   specified command

        // kludge to save flags for "Select DIR ..."
        lSaveDirFlags = glDirFlags;

        // if LFN/HPFS names with whitespace, quote them
        if (glDirFlags & DIRFLAGS_HPFS)
          {
          szFormat = (((strpbrk(sel->sfiles[n].hpfs_name, " \t,+=:") != 0L)
                        || (strpbrk(sel->szPath, " \t,+=|<>") != 0L))
                      ? "%s\"%s%Fs\"%s" : "%s%s%Fs%s");
          sprintf(szCmdline, szFormat, sel->cmd_start, sel->szPath,
                  ((glDirFlags & DIRFLAGS_NT_ALT_NAME)
                  ? sel->sfiles[n].file_name : sel->sfiles[n].hpfs_name),
                  sel->cmd_tail);
          }
        else
          {
          // reassemble pre-/J name
          if ((glDirFlags & DIRFLAGS_JUSTIFY)
              && (sel->sfiles[n].file_name[8] == ' '))
            {
            int j;

            sel->sfiles[n].file_name[8] = '.';
            j = 8;
            while ((j > 0) && (sel->sfiles[n].file_name[j-1] == ' '))
              j--;

            if (j < 8)
              strcpy(sel->sfiles[n].file_name + j,
                     sel->sfiles[n].file_name + 8);
            }

          sprintf(szCmdline, "%s%s%Fs%s", sel->cmd_start,
                  sel->szPath, sel->sfiles[n].file_name, sel->cmd_tail);
          }  // else

        rval = command(szCmdline, 0);

        // restore flags (primarily for DIRFLAGS_HPFS)
        glDirFlags = lSaveDirFlags;
        *szCmdline = '\0';
        }  // if (sel->open_paren == '(')
      else
        {
        // if LFN/HPFS names with whitespace, quote them
        if (glDirFlags & DIRFLAGS_HPFS)
          {
          szFormat = (((strpbrk(sel->sfiles[n].hpfs_name, " \t,+=:") != 0L)
                      || (strpbrk(sel->szPath, " \t,+=:") != 0L))
                      ? "%s\"%Fs\" " : "%s%Fs ");
          sprintf(strend(szCmdline), szFormat, sel->szPath,
                  sel->sfiles[n].hpfs_name);
          }
        else
          sprintf(strend(szCmdline), "%s%Fs ", sel->szPath,
                  sel->sfiles[n].file_name);
        }
      }
    else
      {
      if (++n < sel->entries)
        continue;
      }

    // reset index & marker
    n = 0;
    i++;
    }  // while ((rval != CTRLC) && (i <= sel->marked))

  // check for [] (put everything on a single line) use
  if (szCmdline[0])
    {
    strins(szCmdline, sel->cmd_start);
    strcat(szCmdline, sel->cmd_tail);
    rval = command(szCmdline, 0);
    }

  HoldSignals();
  return rval;
  }

