// XTRACMDS.C - extra commands for 4OS2
// Copyright 2003, Doug Lindquist

#define INCL_DOSPROCESS
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <share.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <errno.h>

#include "product.h"
#include "4all.h"
#include "ps.h"

// used by PS command
typedef struct _plist_node
  {
  qsPrec prec;
  struct _plist_node *next;
  } plist_node;

static plist_node *add_node(plist_node *list, qsPrec *prec);
static CHAR buf[BUFSIZE]; // use 64k buffer as recommended for V2

/**********************************************************************/
/*  make a pointer out of the offset of the pointer passed and the    */
/*  selector of the procstat buffer.                                  */
/*  Not elegant, but utilitarian.                                     */
/**********************************************************************/
static PVOID makeptr(PVOID p)
  {
  ULONG wk1, wk2, wk3;

  wk1 = (ULONG)buf;
  wk2 = (ULONG)p;
  wk3 = (wk1 & 0xffff0000) | (wk2 & 0x0000ffff);
  return (PVOID)wk3;
  }
/************************************************************************/
/*    modname2() - this is how you get the module name in 2.0           */
/************************************************************************/
static PCHAR modname2(USHORT module)
  {
  qsLrec *q;

  q = (qsLrec *)makeptr(((qsPtrrec *)buf)->pLibRec);
  while (q->hmte != module && q->pNextRec)
    q = (qsLrec *)makeptr(q->pNextRec);

  if (q->hmte == module)
    return((PCHAR)makeptr(q->pName));

  return("not found");
  }
/************************************************************************/
/*    qProcStatus()                                                     */
/************************************************************************/
static VOID qProcStatus(BOOL useCondensed)
  {
  qsPrec *pPr;
  qsTrec *pTr;
  PCHAR p;
  INT rc;
  USHORT pid, ppid, hMte, cTCB, cLib, c16Sem, cShrMem;
  ULONG sgid;
  PSZ szCaption[2]=
     {
     "  PID  PPID   SID MHNDL THRDS  MODS  SEMS  SSEG COMMAND",
     "  PID COMMAND"
     };

  rc = DosQProcStatus(buf, sizeof(buf));
  if (rc)
    {
    printf("qProcStatus: DosQProcStatus failed. rc = %d.\n", rc);
//    MessageBox(s, "ERROR");
    return;
    }

  if (useCondensed == TRUE)
    puts(szCaption[1]);
  else
    puts(szCaption[0]);

  pPr = (qsPrec *)makeptr(((qsPtrrec *)buf)->pProcRec);
  while (pPr->RecType == PROCESS_RECORD_TYPE)
    {
    p = modname2(pPr->hMte);
    pid = pPr->pid;
    ppid = pPr->ppid;
    sgid = pPr->sgid;
    hMte = pPr->hMte;
    cTCB = pPr->cTCB;
    cLib = pPr->cLib;
    c16Sem = pPr->c16Sem;
    cShrMem = pPr->cShrMem;
    if (useCondensed == TRUE)
      printf("%5u %s\n", pid, p);
    else
      printf("%5u %5u %5u %5u %5u %5u %5u %5u %s\n",
             pid, ppid, (USHORT)sgid, hMte, cTCB, cLib, c16Sem, cShrMem, p);

    pTr = (qsTrec *)makeptr(pPr->pThrdRec);
    pTr += pPr->cTCB;
    pPr = (qsPrec *)pTr;
    }  // while (pPr->RecType == PROCESS_RECORD_TYPE)
  }
/************************************************************************/
INT ps_cmd(int argc, char **argv)
  {
  PSZ arg;
  INT i=1;
  BOOL useCondensed=TRUE;

  while (i < argc)
    {
    arg = ntharg(*argv, i);
    if ((arg[0] == '-') || (arg[0] == '/'))
      {
      switch (arg[1])
        {
        case 'l':
        case 'L':
          useCondensed = FALSE;
          break;

        case '?':
          puts(PS_USAGE);
          return 1;
        }
      }  // if ((arg[0] == '-') || (arg[0] == '/'))

    ++i;
    }  // while (i < argc)

  qProcStatus(useCondensed);
  return 0;
  }
/************************************************************************/
INT kill_cmd(int argc, char **argv)
  {
  APIRET rc;
  HFILE hfd;
  PID pid;
  ULONG plen, action;
  PSZ arg;

  if (argc == 1)
    return usage(KILL_USAGE);

  arg = ntharg(*argv, 1);
  if (((arg[0] == '-') || (arg[0] == '/')) && (arg[1] == '?'))
    {
    puts(KILL_USAGE);
    return 1;
    }

  if (!isdigit(arg[0]))
    return usage(KILL_USAGE);

  pid = atoi(arg);

  rc = DosKillProcess(DKP_PROCESS, pid);
  if (rc == NO_ERROR)
    return rc;

  rc = DosOpen((PSZ)"/dev/fastio$", (PHFILE)&hfd, (PULONG)&action,
               (ULONG)0, FILE_SYSTEM, FILE_OPEN,
               OPEN_SHARE_DENYNONE|OPEN_FLAGS_NOINHERIT|OPEN_ACCESS_READONLY,
               (ULONG)0);
  if (rc == NO_ERROR)
    {
    rc = DosDevIOCtl(hfd, (ULONG)0x76, (ULONG)0x65, (PULONG*)&pid,
                     sizeof(USHORT), &plen, NULL, 0, NULL);
    if (rc == NO_ERROR)
      {
      DosClose(hfd);
      return rc;
      }
    }

  printf("Unable to kill process %d\n", pid);
  return rc;
  }
/************************************************************************/
INT popups_cmd(INT argc, CHAR **argv)
  {
  CHAR Drive='C';
  PSZ arg, arg1;

  if (argc == 1)
    goto usage;

  arg = ntharg(*argv, 1);
  strupr(arg);

  if (!strcmp(arg, "OFF"))
    {
    if (argc == 3)
      {
      arg1 = ntharg(*argv, 2);
      toupper(arg1[0]);
      Drive = arg1[0];
      if ((Drive < 'A') || ((Drive > 'Z') && (Drive < 'a')) || (Drive > 'z'))
        Drive = 'C';
      }

    DosSuppressPopUps(SPU_ENABLESUPPRESSION, Drive);
    puts("Popup suppression is now turned on");
    return 0;
    }

  if (!strcmp(arg, "ON"))
    {
    DosSuppressPopUps(SPU_DISABLESUPPRESSION, 0);
    puts("Popup suppression is now turned off");
    return 0;
    }

usage:
  puts("POPUPS [OFF | ON [drive]]\n");
  return 1;
  }

